import pytest

from eth_utils import (
    is_hexstr,
)


@pytest.mark.parametrize(
    "value,expected",
    (
        ("", False),  # empty string
        ("0x", True),  # empty string
        ("0X", True),  # empty string
        ("abcdef1234567890", True),
        ("ABCDEF1234567890", True),
        ("AbCdEf1234567890", True),
        ("0xabcdef1234567890", True),
        ("0xABCDEF1234567890", True),
        ("0xAbCdEf1234567890", True),
        ("12345", True),  # odd length
        ("0x12345", True),  # odd length
        ("123456xx", False),  # non-hex character
        ("0x123456xx", False),  # non-hex character
        ("0\u0080", False),  # triggers different exceptions in py2 and py3
        (1, False),  # int
        (b"", False),  # bytes
        ({}, False),  # dictionary
        (lambda: None, False),  # lambda function
    ),
)
def test_is_hexstr(value, expected):
    actual = is_hexstr(value)
    assert actual is expected



from dataclasses import (
    dataclass,
)
import json
import os
from typing import (
    List,
)

from eth_typing import (
    ChainId,
)

from eth_utils import (
    ValidationError,
)


@dataclass
class Network:
    chain_id: int
    name: str
    shortName: str
    symbol: ChainId


def initialize_network_objects() -> List[Network]:
    networks_obj = []

    networks_json_path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "__json")
    )
    with open(
        os.path.join(networks_json_path, "eth_networks.json"),
        encoding="UTF-8",
    ) as open_file:
        network_data = json.load(open_file)

    for entry in network_data:
        try:
            network = Network(
                chain_id=entry["chainId"],
                name=entry["name"],
                shortName=entry["shortName"],
                symbol=ChainId(entry["chainId"]),
            )
            networks_obj.append(network)
        except ValueError:
            # Chain does not have a valid ChainId, network files in eth-utils and
            # eth-typing should to be updated. Run `python update_networks.py` in the
            # project root.
            pass

    return networks_obj


networks = initialize_network_objects()

networks_by_id = {network.chain_id: network for network in networks}
network_names_by_id = {network.chain_id: network.name for network in networks}
network_short_names_by_id = {
    network.chain_id: network.shortName for network in networks
}


def network_from_chain_id(chain_id: int) -> Network:
    try:
        return networks_by_id[chain_id]
    except KeyError:
        raise ValidationError(f"chain_id is not recognized: {chain_id}")


def name_from_chain_id(chain_id: int) -> str:
    try:
        return network_names_by_id[chain_id]
    except KeyError:
        raise ValidationError(f"chain_id is not recognized: {chain_id}")


def short_name_from_chain_id(chain_id: int) -> str:
    try:
        return network_short_names_by_id[chain_id]
    except KeyError:
        raise ValidationError(f"chain_id is not recognized: {chain_id}")



import pytest

from eth_typing import (
    ChainId,
)

from eth_utils import (
    Network,
    ValidationError,
    name_from_chain_id,
    network_from_chain_id,
    short_name_from_chain_id,
)


@pytest.mark.parametrize(
    ("chain_id", "expected_name"),
    [
        (1, "Ethereum Mainnet"),
        (2, "Expanse Network"),
        (42, "LUKSO Mainnet"),
    ],
)
def test_network_name_from_chain_id(chain_id, expected_name):
    result = name_from_chain_id(chain_id)
    assert result == expected_name


@pytest.mark.parametrize(
    ("chain_id", "expected_short_name"),
    [
        (1, "eth"),
        (2, "exp"),
        (42, "lukso"),
    ],
)
def test_short_name_from_chain_id(chain_id, expected_short_name):
    result = short_name_from_chain_id(chain_id)
    assert isinstance(result, str)
    assert result == expected_short_name


@pytest.mark.parametrize(
    ("chain_id", "expected_network"),
    [
        (1, Network(1, "Ethereum Mainnet", "eth", ChainId.ETH)),
        (2, Network(2, "Expanse Network", "exp", ChainId.EXP)),
        (42, Network(42, "LUKSO Mainnet", "lukso", ChainId.LUKSO)),
        (43, Network(43, "Darwinia Pangolin Testnet", "pangolin", ChainId.PANGOLIN)),
    ],
)
def test_network_from_chain_id(chain_id, expected_network):
    result = network_from_chain_id(chain_id)
    assert isinstance(result, Network)
    assert result == expected_network


@pytest.mark.parametrize("invalid_chain_id", (222**222, 333**333, 444**444))
def test_from_chain_id_utility_methods_with_invalid_chain_id(invalid_chain_id):
    error_message = f"chain_id is not recognized: {invalid_chain_id}"

    # test network_from_chain_id
    with pytest.raises(ValidationError, match=error_message):
        network_from_chain_id(invalid_chain_id)

    # test short_name_from_chain_id
    with pytest.raises(ValidationError, match=error_message):
        short_name_from_chain_id(invalid_chain_id)

    # test name_from_chain_id
    with pytest.raises(ValidationError, match=error_message):
        name_from_chain_id(invalid_chain_id)



import pytest

from eth_utils.hexadecimal import (
    add_0x_prefix,
    remove_0x_prefix,
)


@pytest.mark.parametrize(
    "value,expected",
    (
        ("", "0x"),
        ("0x", "0x"),
        ("0x12345", "0x12345"),
        ("0X12345", "0X12345"),
        ("12345", "0x12345"),
    ),
)
def test_add_0x_prefix(value, expected):
    actual = add_0x_prefix(value)
    assert actual == expected


@pytest.mark.parametrize("value", (b"", 123, {}, lambda: None))
def test_add_0x_prefix_rejects_non_text_types(value):
    with pytest.raises(TypeError):
        add_0x_prefix(value)


@pytest.mark.parametrize(
    "value,expected",
    (
        ("", ""),
        ("0x", ""),
        ("0x12345", "12345"),
        ("0X12345", "12345"),
        ("12345", "12345"),
    ),
)
def test_remove_0x_prefix(value, expected):
    actual = remove_0x_prefix(value)
    assert actual == expected


@pytest.mark.parametrize("value", (b"", 123, {}, lambda: None))
def test_remove_0x_prefix_rejects_non_text_types(value):
    with pytest.raises(TypeError):
        remove_0x_prefix(value)



from typing import (
    List,
    Tuple,
)

from eth_utils import (
    to_ordered_dict,
)


@to_ordered_dict
def return_value() -> List[Tuple[int, int]]:
    return [(1, 2)]


x = return_value()
reveal_type(x)  # noqa: F821



from typing import (
    List,
)

from eth_utils.curried import (  # noqa: F401
    ExtendedDebugLogger,
    HasExtendedDebugLogger,
    HasExtendedDebugLoggerMeta,
    HasLogger,
    HasLoggerMeta,
    ValidationError,
    add_0x_prefix,
    apply_formatter_at_index,
    apply_formatter_if,
    apply_formatter_to_array,
    apply_formatters_to_dict,
    apply_formatters_to_sequence,
    apply_key_map,
    apply_one_of_formatters,
    apply_to_return_value,
    big_endian_to_int,
    clamp,
    combine_argument_formatters,
    combomethod,
    decode_hex,
    denoms,
    encode_hex,
    event_abi_to_log_topic,
    event_signature_to_log_topic,
    flatten_return,
    from_wei,
    function_abi_to_4byte_selector,
    function_signature_to_4byte_selector,
    get_extended_debug_logger,
    get_logger,
    hexstr_if_str,
    humanize_bytes,
    humanize_hash,
    humanize_integer_sequence,
    humanize_ipfs_uri,
    humanize_seconds,
    humanize_wei,
    import_string,
    int_to_big_endian,
    is_0x_prefixed,
    is_address,
    is_binary_address,
    is_boolean,
    is_bytes,
    is_canonical_address,
    is_checksum_address,
    is_checksum_formatted_address,
    is_dict,
    is_hex,
    is_hex_address,
    is_hexstr,
    is_integer,
    is_list,
    is_list_like,
    is_normalized_address,
    is_null,
    is_number,
    is_same_address,
    is_string,
    is_text,
    is_tuple,
    keccak,
    remove_0x_prefix,
    replace_exceptions,
    reversed_return,
    setup_DEBUG2_logging,
    sort_return,
    text_if_str,
    to_bytes,
    to_canonical_address,
    to_checksum_address,
    to_dict,
    to_hex,
    to_int,
    to_list,
    to_normalized_address,
    to_ordered_dict,
    to_set,
    to_text,
    to_tuple,
    to_wei,
)


def i_put_my_thing_down_flip_it_and_reverse_it(lyric: List[str]) -> str:
    return "".join(reversed(lyric))


CONDITION_FORMATTER_PAIRS = (
    (is_list_like, tuple),
    (is_string, i_put_my_thing_down_flip_it_and_reverse_it),
)


apply_formatter_if(is_string)
apply_formatter_if(is_string, bool)
apply_formatter_if(is_string, bool, 1)


apply_one_of_formatters(CONDITION_FORMATTER_PAIRS)
apply_one_of_formatters(CONDITION_FORMATTER_PAIRS, "my thing")



import pytest

from eth_utils import (
    to_bytes,
    to_hex,
    to_int,
    to_text,
)


@pytest.mark.parametrize(
    "val, expected",
    (
        (b"\x01", b"\x01"),
        (b"\xff", b"\xff"),
        (b"\x00", b"\x00"),
        (bytearray(b"\x01"), b"\x01"),
        (bytearray(b"\xff"), b"\xff"),
        (bytearray(b"\x00"), b"\x00"),
        (memoryview(b"\x01"), b"\x01"),
        (memoryview(b"\xff"), b"\xff"),
        (memoryview(b"\x00"), b"\x00"),
        (0x1, b"\x01"),
        (0x0001, b"\x01"),
        (0xFF, b"\xff"),
        (0, b"\x00"),
        (256, b"\x01\x00"),
        (True, b"\x01"),
        (False, b"\x00"),
    ),
)
def test_to_bytes_primitive(val, expected):
    assert to_bytes(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        ("0x", b""),
        ("0x0", b"\x00"),
        ("0x1", b"\x01"),
        ("0", b"\x00"),
        ("1", b"\x01"),
        ("0xFF", b"\xff"),
        ("0x100", b"\x01\x00"),
        ("0x0000", b"\x00\x00"),
        ("0000", b"\x00\x00"),
    ),
)
def test_to_bytes_hexstr(val, expected):
    assert to_bytes(hexstr=val) == expected


@pytest.mark.parametrize("val, expected", (("cowmö", b"cowm\xc3\xb6"), ("", b"")))
def test_to_bytes_text(val, expected):
    assert to_bytes(text=val) == expected


def test_to_text_identity():
    assert to_text(text="pass-through") == "pass-through"


@pytest.mark.parametrize(
    "val, expected",
    (
        (b"", ""),
        (bytearray(b""), ""),
        (memoryview(b""), ""),
        ("0x", ""),
        (b"cowm\xc3\xb6", "cowmö"),
        (bytearray(b"cowm\xc3\xb6"), "cowmö"),
        (memoryview(b"cowm\xc3\xb6"), "cowmö"),
        ("0x636f776dc3b6", "cowmö"),
        (0x636F776DC3B6, "cowmö"),
        ("0xa", "\n"),
    ),
)
def test_to_text(val, expected):
    assert to_text(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (("0x", ""), ("0xa", "\n"), ("0x636f776dc3b6", "cowmö"), ("636f776dc3b6", "cowmö")),
)
def test_to_text_hexstr(val, expected):
    assert to_text(hexstr=val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (b"\x00", 0),
        (b"\x01", 1),
        (b"\x00\x01", 1),
        (b"\x01\x00", 256),
        (bytearray(b"\x00"), 0),
        (bytearray(b"\x01"), 1),
        (bytearray(b"\x00\x01"), 1),
        (bytearray(b"\x01\x00"), 256),
        (memoryview(b"\x01"), 1),
        (memoryview(b"\xff"), 255),
        (memoryview(b"\x00"), 0),
        (True, 1),
        (False, 0),
        ("255", TypeError),
        ("-1", TypeError),
        ("0x0", TypeError),
        ("0x1", TypeError),
    ),
)
def test_to_int(val, expected):
    if isinstance(expected, type):
        with pytest.raises(expected):
            to_int(val)
    else:
        assert to_int(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        ("0", 0),
        ("-1", -1),
        ("255", 255),
        ("0x0", ValueError),
        ("0x1", ValueError),
        ("1.1", ValueError),
        ("a", ValueError),
    ),
)
def test_to_int_text(val, expected):
    if isinstance(expected, type):
        with pytest.raises(expected):
            to_int(text=val)
    else:
        assert to_int(text=val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        ("0x0", 0),
        ("0x1", 1),
        ("0x01", 1),
        ("0x10", 16),
        ("0", 0),
        ("1", 1),
        ("01", 1),
        ("10", 16),
    ),
)
def test_to_int_hexstr(val, expected):
    assert to_int(hexstr=val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (b"\x00", "0x00"),
        (b"\x01", "0x01"),
        (b"\x10", "0x10"),
        (b"\x01\x00", "0x0100"),
        (b"\x00\x0F", "0x000f"),
        (b"", "0x"),
        (bytearray(b"\x00"), "0x00"),
        (bytearray(b"\x01"), "0x01"),
        (bytearray(b"\x10"), "0x10"),
        (bytearray(b"\x01\x00"), "0x0100"),
        (bytearray(b"\x00\x0F"), "0x000f"),
        (bytearray(b""), "0x"),
        (memoryview(b"\x00"), "0x00"),
        (memoryview(b"\x01"), "0x01"),
        (memoryview(b"\x10"), "0x10"),
        (memoryview(b"\x01\x00"), "0x0100"),
        (memoryview(b"\x00\x0F"), "0x000f"),
        (memoryview(b""), "0x"),
        (0, "0x0"),
        (1, "0x1"),
        (16, "0x10"),
        (256, "0x100"),
        (0x0, "0x0"),
        (0x0F, "0xf"),
        (False, "0x0"),
        (True, "0x1"),
    ),
)
def test_to_hex(val, expected):
    assert to_hex(val) == expected


@pytest.mark.parametrize("val, expected", (("", "0x"), ("cowmö", "0x636f776dc3b6")))
def test_to_hex_text(val, expected):
    assert to_hex(text=val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        ("0x0", "0x0"),
        ("0x1", "0x1"),
        ("0x0001", "0x0001"),
        ("0x10", "0x10"),
        ("0xF", "0xf"),
        ("F", "0xf"),
    ),
)
def test_to_hex_cleanup_only(val, expected):
    assert to_hex(hexstr=val) == expected


def test_to_hex_exception():
    with pytest.raises(TypeError, match="Unsupported type: The primitive argument"):
        to_hex(primitive="string")



import pytest

from eth_utils.types import (
    is_boolean,
    is_dict,
    is_integer,
    is_list,
    is_list_like,
    is_string,
    is_tuple,
)


@pytest.mark.parametrize(
    "value,expected",
    [
        (3, True),
        (None, False),
        ("3", False),
        ("0x3", False),
        (True, False),
        (False, False),
    ],
)
def test_is_integer(value, expected):
    assert is_integer(value) == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        ("3", True),
        (b"3", True),
        (b"", True),
        ("", True),
        (None, False),
        (3, False),
        ({}, False),
    ],
)
def test_is_string(value, expected):
    assert is_string(value) == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        (lambda: None, False),
        (3, False),
        (None, False),
        ("3", False),
        ("0x3", False),
        ({}, True),
        ({"test": 3}, True),
    ],
)
def test_is_dict(value, expected):
    assert is_dict(value) == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        (lambda: None, False),
        (3, False),
        (None, False),
        ("3", False),
        ("0x3", False),
        (True, True),
        (False, True),
    ],
)
def test_is_boolean(value, expected):
    assert is_boolean(value) == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        (lambda: None, False),
        (3, False),
        (None, False),
        ("3", False),
        ("0x3", False),
        ([], True),
        ([3], True),
        ([3, 3], True),
        ([None], True),
        ([tuple()], True),
        ([(1, 2)], True),
        (tuple(), True),
        ((3,), True),
        ((3, 3), True),
        ((None,), True),
        ((tuple(),), True),
        (((1, 2),), True),
    ],
)
def test_is_list_like(value, expected):
    assert is_list_like(value) == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        (lambda: None, False),
        (3, False),
        (None, False),
        ("3", False),
        ("0x3", False),
        ([], True),
        ([3], True),
        ([3, 3], True),
        ([None], True),
        ([tuple()], True),
        ([(1, 2)], True),
        (tuple(), False),
        ((3,), False),
        ((3, 3), False),
        ((None,), False),
        ((tuple(),), False),
        (((1, 2),), False),
    ],
)
def test_is_list(value, expected):
    assert is_list(value) == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        (lambda: None, False),
        (3, False),
        (None, False),
        ("3", False),
        ("0x3", False),
        ([], False),
        ([3], False),
        ([3, 3], False),
        ([None], False),
        ([tuple()], False),
        ([(1, 2)], False),
        (tuple(), True),
        ((3,), True),
        ((3, 3), True),
        ((None,), True),
        ((tuple(),), True),
        (((1, 2),), True),
    ],
)
def test_is_tuple(value, expected):
    assert is_tuple(value) == expected



import logging

from eth_utils.logging import (
    ExtendedDebugLogger,
    HasExtendedDebugLogger,
    HasLogger,
    HasLoggerMeta,
)


class DefinedAtModule(HasLogger):
    pass


def test_simple_class_has_logger():
    assert hasattr(DefinedAtModule, "logger")
    assert isinstance(DefinedAtModule.logger, logging.Logger)

    assert "DefinedAtModule" in DefinedAtModule.logger.name


def test_has_logger_for_class_defined_in_function_scope():
    class ForTest(HasLogger):
        pass

    assert hasattr(ForTest, "logger")
    assert isinstance(ForTest.logger, logging.Logger)

    assert "ForTest" in ForTest.logger.name
    assert "test_has_logger_for_class_defined_in_function_scope" in ForTest.logger.name


class OuterClass:
    class InnerClass(HasLogger):
        pass


def test_nested_class_has_logger():
    assert hasattr(OuterClass.InnerClass, "logger")
    assert isinstance(OuterClass.InnerClass.logger, logging.Logger)

    assert "OuterClass.InnerClass" in OuterClass.InnerClass.logger.name


class ExplicitLogger(HasLogger):
    logger = logging.getLogger("explicitely-declared")


def test_has_logger_respects_existing_logger():
    assert ExplicitLogger.logger.name == "explicitely-declared"


class Parent(HasLogger):
    pass


class Child(Parent):
    pass


def test_inherited_gets_own_with_has_logger():
    assert hasattr(Parent, "logger")
    assert hasattr(Child, "logger")

    assert Parent.logger.name.endswith("Parent")
    assert Child.logger.name.endswith("Child")
    assert "Parent" not in Child.logger.name


class InheritsExplicit(ExplicitLogger):
    pass


def test_inherited_from_explicit_gets_own_logger_with_has_logger():
    assert hasattr(InheritsExplicit, "logger")
    assert InheritsExplicit.logger.name.endswith("InheritsExplicit")
    assert "ExplicitLogger" not in InheritsExplicit.logger.name


class FancyLogger(logging.Logger):
    pass


def test_can_override_logging_class():
    class HasFancyLogger(metaclass=HasLoggerMeta.replace_logger_class(FancyLogger)):
        pass

    assert hasattr(HasFancyLogger, "logger")
    assert isinstance(HasFancyLogger.logger, FancyLogger)
    assert "HasFancyLogger" in HasFancyLogger.logger.name


def test_extended_debug_logger():
    assert isinstance(HasExtendedDebugLogger.logger, ExtendedDebugLogger)
    assert "HasExtendedDebugLogger" in HasExtendedDebugLogger.logger.name
    assert "_BaseHasExtendedDebugLogger" not in HasExtendedDebugLogger.logger.name


class InheritsExtended(HasExtendedDebugLogger):
    pass


def test_inherited_has_extended_debug_logger():
    assert hasattr(InheritsExtended, "logger")
    assert isinstance(InheritsExtended.logger, ExtendedDebugLogger)
    assert InheritsExtended.logger.name.endswith("InheritsExtended")



import pytest

from eth_utils.numeric import (
    clamp,
)


@pytest.mark.parametrize(
    "lower_bound,upper_bound,value,expected",
    ((5, 7, 4, 5), (5, 7, 5, 5), (5, 7, 6, 6), (5, 7, 7, 7), (5, 7, 8, 7)),
)
def test_numeric_clamp_utility(lower_bound, upper_bound, value, expected):
    result = clamp(lower_bound, upper_bound, value)
    assert result == expected



from collections import (
    abc,
)
import copy
import itertools
import re
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
    overload,
)

from eth_typing import (
    ABI,
    ABIComponent,
    ABIConstructor,
    ABIElement,
    ABIError,
    ABIEvent,
    ABIFallback,
    ABIFunction,
    ABIReceive,
)

from eth_utils.types import (
    is_list_like,
)

from .crypto import (
    keccak,
)


def _align_abi_input(
    arg_abi: ABIComponent, normalized_arg: Any
) -> Union[Any, Tuple[Any, ...]]:
    """
    Aligns the values of any mapping at any level of nesting in ``normalized_arg``
    according to the layout of the corresponding abi spec.
    """
    tuple_parts = _get_tuple_type_str_and_dims(arg_abi.get("type", ""))

    if tuple_parts is None:
        # normalized_arg is non-tuple.  Just return value.
        return normalized_arg

    tuple_prefix, tuple_dims = tuple_parts
    if tuple_dims is None:
        # normalized_arg is non-list tuple.  Each sub arg in `normalized_arg` will be
        # aligned according to its corresponding abi.
        sub_abis = cast(Iterable[ABIComponent], arg_abi.get("components", []))
    else:
        num_dims = tuple_dims.count("[")

        # normalized_arg is list tuple.  A non-list version of its abi will be used to
        # align each element in `normalized_arg`.
        new_abi = copy.copy(arg_abi)
        new_abi["type"] = tuple_prefix + "[]" * (num_dims - 1)

        sub_abis = itertools.repeat(new_abi)

    if isinstance(normalized_arg, abc.Mapping):
        # normalized_arg is mapping.  Align values according to abi order.
        aligned_arg = tuple(normalized_arg[abi["name"]] for abi in sub_abis)
    else:
        aligned_arg = normalized_arg

    if not is_list_like(aligned_arg):
        raise TypeError(
            f'Expected non-string sequence for "{arg_abi.get("type")}" '
            f"component type: got {aligned_arg}"
        )

    # convert NamedTuple to regular tuple
    typing = tuple if isinstance(aligned_arg, tuple) else type(aligned_arg)

    return typing(
        _align_abi_input(sub_abi, sub_arg)
        for sub_abi, sub_arg in zip(sub_abis, aligned_arg)
    )


def _get_tuple_type_str_and_dims(s: str) -> Optional[Tuple[str, Optional[str]]]:
    """
    Takes a JSON ABI type string.  For tuple type strings, returns the separated
    prefix and array dimension parts.  For all other strings, returns ``None``.
    """
    tuple_type_str_re = "^(tuple)((\\[([1-9]\\d*\b)?])*)??$"
    match = re.compile(tuple_type_str_re).match(s)

    if match is not None:
        tuple_prefix = match.group(1)
        tuple_dims = match.group(2)

        return tuple_prefix, tuple_dims

    return None


def _raise_if_not_function_abi(abi_element: ABIElement) -> None:
    if abi_element["type"] != "function":
        raise ValueError(
            f"Outputs only supported for ABI type `function`. Provided"
            f" ABI type was `{abi_element.get('type')}` and outputs were "
            f"`{abi_element.get('outputs')}`."
        )


def _raise_if_fallback_or_receive_abi(abi_element: ABIElement) -> None:
    if abi_element["type"] == "fallback" or abi_element["type"] == "receive":
        raise ValueError(
            f"Inputs not supported for function types `fallback` or `receive`. Provided"
            f" ABI type was `{abi_element.get('type')}` with inputs "
            f"`{abi_element.get('inputs')}`."
        )


def collapse_if_tuple(abi: Union[ABIComponent, Dict[str, Any], str]) -> str:
    """
    Extract argument types from a function or event ABI parameter.

    With tuple argument types, return a Tuple of each type.
    Returns the param if `abi` is an instance of str or another non-tuple
    type.

    :param abi: A Function or Event ABI component or a string with type info.
    :type abi: `Union[ABIComponent, Dict[str, Any], str]`
    :return: Type(s) for the function or event ABI param.
    :rtype: `str`

    .. doctest::

        >>> from eth_utils.abi import collapse_if_tuple
        >>> abi = {
        ...   'components': [
        ...     {'name': 'anAddress', 'type': 'address'},
        ...     {'name': 'anInt', 'type': 'uint256'},
        ...     {'name': 'someBytes', 'type': 'bytes'},
        ...   ],
        ...   'type': 'tuple',
        ... }
        >>> collapse_if_tuple(abi)
        '(address,uint256,bytes)'
    """
    if isinstance(abi, str):
        return abi

    element_type = abi.get("type")
    if not isinstance(element_type, str):
        raise TypeError(
            f"The 'type' must be a string, but got {repr(element_type)} of type "
            f"{type(element_type)}"
        )
    elif not element_type.startswith("tuple"):
        return element_type

    delimited = ",".join(collapse_if_tuple(c) for c in abi["components"])
    # Whatever comes after "tuple" is the array dims. The ABI spec states that
    # this will have the form "", "[]", or "[k]".
    array_dim = element_type[5:]
    collapsed = f"({delimited}){array_dim}"

    return collapsed


def abi_to_signature(abi_element: ABIElement) -> str:
    """
    Returns a string signature representation of the function or event ABI
    and arguments.

    Signatures consist of the name followed by a list of arguments.

    :param abi_element: ABI element.
    :type abi_element: `ABIElement`
    :return: Stringified ABI signature
    :rtype: `str`

    .. doctest::

        >>> from eth_utils import abi_to_signature
        >>> abi_element = {
        ...   'constant': False,
        ...   'inputs': [
        ...     {
        ...       'name': 's',
        ...       'type': 'uint256'
        ...     }
        ...   ],
        ...   'name': 'f',
        ...   'outputs': [],
        ...   'payable': False,
        ...   'stateMutability': 'nonpayable',
        ...   'type': 'function'
        ... }
        >>> abi_to_signature(abi_element)
        'f(uint256)'
    """
    signature = "{name}({input_types})"

    abi_type = str(abi_element.get("type", ""))
    if abi_type == "fallback" or abi_type == "receive":
        return signature.format(name=abi_type, input_types="")

    if abi_type == "constructor":
        fn_name = abi_type
    else:
        fn_name = str(abi_element.get("name", abi_type))

    return signature.format(
        name=fn_name, input_types=",".join(get_abi_input_types(abi_element))
    )


def filter_abi_by_name(abi_name: str, contract_abi: ABI) -> Sequence[ABIElement]:
    """
    Get one or more function and event ABIs by name.

    :param abi_name: Name of the function, event or error.
    :type abi_name: `str`
    :param contract_abi: Contract ABI.
    :type contract_abi: `ABI`
    :return: Function or event ABIs with matching name.
    :rtype: `Sequence[ABIElement]`

    .. doctest::

            >>> from eth_utils.abi import filter_abi_by_name
            >>> abi = [
            ...     {
            ...         "constant": False,
            ...         "inputs": [],
            ...         "name": "func_1",
            ...         "outputs": [],
            ...         "type": "function",
            ...     },
            ...     {
            ...         "constant": False,
            ...         "inputs": [
            ...             {"name": "a", "type": "uint256"},
            ...         ],
            ...         "name": "func_2",
            ...         "outputs": [],
            ...         "type": "function",
            ...     },
            ...     {
            ...         "constant": False,
            ...         "inputs": [
            ...             {"name": "a", "type": "uint256"},
            ...             {"name": "b", "type": "uint256"},
            ...         ],
            ...         "name": "func_3",
            ...         "outputs": [],
            ...         "type": "function",
            ...     },
            ...     {
            ...         "constant": False,
            ...         "inputs": [
            ...             {"name": "a", "type": "uint256"},
            ...             {"name": "b", "type": "uint256"},
            ...             {"name": "c", "type": "uint256"},
            ...         ],
            ...         "name": "func_4",
            ...         "outputs": [],
            ...         "type": "function",
            ...     },
            ... ]
            >>> filter_abi_by_name("func_1", abi)
            [{'constant': False, 'inputs': [], 'name': 'func_1', 'outputs': [], \
'type': 'function'}]
    """
    return [
        abi
        for abi in contract_abi
        if (
            (
                abi["type"] == "function"
                or abi["type"] == "event"
                or abi["type"] == "error"
            )
            and abi["name"] == abi_name
        )
    ]


@overload
def filter_abi_by_type(
    abi_type: Literal["function"],
    contract_abi: ABI,
) -> Sequence[ABIFunction]:
    pass


@overload
def filter_abi_by_type(
    abi_type: Literal["constructor"],
    contract_abi: ABI,
) -> Sequence[ABIConstructor]:
    pass


@overload
def filter_abi_by_type(
    abi_type: Literal["fallback"],
    contract_abi: ABI,
) -> Sequence[ABIFallback]:
    pass


@overload
def filter_abi_by_type(
    abi_type: Literal["receive"],
    contract_abi: ABI,
) -> Sequence[ABIReceive]:
    pass


@overload
def filter_abi_by_type(
    abi_type: Literal["event"],
    contract_abi: ABI,
) -> Sequence[ABIEvent]:
    pass


@overload
def filter_abi_by_type(
    abi_type: Literal["error"],
    contract_abi: ABI,
) -> Sequence[ABIError]:
    pass


def filter_abi_by_type(
    abi_type: Literal[
        "function", "constructor", "fallback", "receive", "event", "error"
    ],
    contract_abi: ABI,
) -> Sequence[
    Union[ABIFunction, ABIConstructor, ABIFallback, ABIReceive, ABIEvent, ABIError]
]:
    """
    Return a list of each ``ABIElement`` that is of type ``abi_type``.

    For mypy, function overloads ensures the correct type is returned based on the
    ``abi_type``. For example, if ``abi_type`` is "function", the return type will be
    ``Sequence[ABIFunction]``.

    :param abi_type: Type of ABI element to filter by.
    :type abi_type: `str`
    :param contract_abi: Contract ABI.
    :type contract_abi: `ABI`
    :return: List of ABI elements of the specified type.
    :rtype: `Sequence[Union[ABIFunction, ABIConstructor, ABIFallback, ABIReceive, \
ABIEvent, ABIError]]`

    .. doctest::

        >>> from eth_utils import filter_abi_by_type
        >>> abi = [
        ...   {"type": "function", "name": "myFunction", "inputs": [], "outputs": []},
        ...   {"type": "function", "name": "myFunction2", "inputs": [], "outputs": []},
        ...   {"type": "event", "name": "MyEvent", "inputs": []}
        ... ]
        >>> filter_abi_by_type("function", abi)
        [{'type': 'function', 'name': 'myFunction', 'inputs': [], 'outputs': []}, \
{'type': 'function', 'name': 'myFunction2', 'inputs': [], 'outputs': []}]
    """
    if abi_type == Literal["function"] or abi_type == "function":
        return [abi for abi in contract_abi if abi["type"] == "function"]
    elif abi_type == Literal["constructor"] or abi_type == "constructor":
        return [abi for abi in contract_abi if abi["type"] == "constructor"]
    elif abi_type == Literal["fallback"] or abi_type == "fallback":
        return [abi for abi in contract_abi if abi["type"] == "fallback"]
    elif abi_type == Literal["receive"] or abi_type == "receive":
        return [abi for abi in contract_abi if abi["type"] == "receive"]
    elif abi_type == Literal["event"] or abi_type == "event":
        return [abi for abi in contract_abi if abi["type"] == "event"]
    elif abi_type == Literal["error"] or abi_type == "error":
        return [abi for abi in contract_abi if abi["type"] == "error"]
    else:
        raise ValueError(f"Unsupported ABI type: {abi_type}")


def get_all_function_abis(contract_abi: ABI) -> Sequence[ABIFunction]:
    """
    Return interfaces for each function in the contract ABI.

    :param contract_abi: Contract ABI.
    :type contract_abi: `ABI`
    :return: List of ABIs for each function interface.
    :rtype: `Sequence[ABIFunction]`

    .. doctest::

        >>> from eth_utils import get_all_function_abis
        >>> contract_abi = [
        ...   {"type": "function", "name": "myFunction", "inputs": [], "outputs": []},
        ...   {"type": "function", "name": "myFunction2", "inputs": [], "outputs": []},
        ...   {"type": "event", "name": "MyEvent", "inputs": []}
        ... ]
        >>> get_all_function_abis(contract_abi)
        [{'type': 'function', 'name': 'myFunction', 'inputs': [], 'outputs': []}, \
{'type': 'function', 'name': 'myFunction2', 'inputs': [], 'outputs': []}]
    """
    return filter_abi_by_type("function", contract_abi)


def get_all_event_abis(contract_abi: ABI) -> Sequence[ABIEvent]:
    """
    Return interfaces for each event in the contract ABI.

    :param contract_abi: Contract ABI.
    :type contract_abi: `ABI`
    :return: List of ABIs for each event interface.
    :rtype: `Sequence[ABIEvent]`

    .. doctest::

        >>> from eth_utils import get_all_event_abis
        >>> contract_abi = [
        ...   {"type": "function", "name": "myFunction", "inputs": [], "outputs": []},
        ...   {"type": "function", "name": "myFunction2", "inputs": [], "outputs": []},
        ...   {"type": "event", "name": "MyEvent", "inputs": []}
        ... ]
        >>> get_all_event_abis(contract_abi)
        [{'type': 'event', 'name': 'MyEvent', 'inputs': []}]
    """
    return filter_abi_by_type("event", contract_abi)


def get_normalized_abi_inputs(
    abi_element: ABIElement,
    *args: Optional[Sequence[Any]],
    **kwargs: Optional[Dict[str, Any]],
) -> Tuple[Any, ...]:
    r"""
    Flattens positional args (``args``) and keyword args (``kwargs``) into a Tuple and
    uses the ``abi_element`` for validation.

    Checks to ensure that the correct number of args were given, no duplicate args were
    given, and no unknown args were given.  Returns a list of argument values aligned
    to the order of inputs defined in ``abi_element``.

    :param abi_element: ABI element.
    :type abi_element: `ABIElement`
    :param args: Positional arguments for the function.
    :type args: `Optional[Sequence[Any]]`
    :param kwargs: Keyword arguments for the function.
    :type kwargs: `Optional[Dict[str, Any]]`
    :return: Arguments list.
    :rtype: `Tuple[Any, ...]`

    .. doctest::

        >>> from eth_utils import get_normalized_abi_inputs
        >>> abi = {
        ...   'constant': False,
        ...   'inputs': [
        ...     {
        ...       'name': 'name',
        ...       'type': 'string'
        ...     },
        ...     {
        ...       'name': 's',
        ...       'type': 'uint256'
        ...     },
        ...     {
        ...       'name': 't',
        ...       'components': [
        ...         {'name': 'anAddress', 'type': 'address'},
        ...         {'name': 'anInt', 'type': 'uint256'},
        ...         {'name': 'someBytes', 'type': 'bytes'},
        ...       ],
        ...       'type': 'tuple'
        ...     }
        ...   ],
        ...   'name': 'f',
        ...   'outputs': [],
        ...   'payable': False,
        ...   'stateMutability': 'nonpayable',
        ...   'type': 'function'
        ... }
        >>> get_normalized_abi_inputs(
        ...   abi, *('myName', 123), **{'t': ('0x1', 1, b'\x01')}
        ... )
        ('myName', 123, ('0x1', 1, b'\x01'))
    """
    _raise_if_fallback_or_receive_abi(abi_element)

    function_inputs = cast(Sequence[ABIComponent], abi_element.get("inputs", []))
    if len(args) + len(kwargs) != len(function_inputs):
        raise TypeError(
            f"Incorrect argument count. Expected '{len(function_inputs)}'"
            f", got '{len(args) + len(kwargs)}'."
        )

    # If no keyword args were given, we don't need to align them
    if not kwargs:
        return cast(Tuple[Any, ...], args)

    kwarg_names = set(kwargs.keys())
    sorted_arg_names = tuple(arg_abi["name"] for arg_abi in function_inputs)
    args_as_kwargs = dict(zip(sorted_arg_names, args))

    # Check for duplicate args
    duplicate_args = kwarg_names.intersection(args_as_kwargs.keys())
    if duplicate_args:
        raise TypeError(
            f"{abi_element.get('name')}() got multiple values for argument(s) "
            f"'{', '.join(duplicate_args)}'."
        )

    # Check for unknown args
    # Arg names sorted to raise consistent error messages
    unknown_args = tuple(sorted(kwarg_names.difference(sorted_arg_names)))
    if unknown_args:
        message = "{} got unexpected keyword argument(s) '{}'."
        if abi_element.get("name"):
            raise TypeError(
                message.format(f"{abi_element.get('name')}()", ", ".join(unknown_args))
            )
        raise TypeError(
            message.format(
                f"Type: '{abi_element.get('type')}'", ", ".join(unknown_args)
            )
        )

    # Sort args according to their position in the ABI and unzip them from their
    # names
    sorted_args = tuple(
        zip(
            *sorted(
                itertools.chain(kwargs.items(), args_as_kwargs.items()),
                key=lambda kv: sorted_arg_names.index(kv[0]),
            )
        )
    )

    if len(sorted_args) > 0:
        return tuple(sorted_args[1])
    else:
        return tuple()


def get_aligned_abi_inputs(
    abi_element: ABIElement,
    normalized_args: Union[Tuple[Any, ...], Mapping[Any, Any]],
) -> Tuple[Tuple[str, ...], Tuple[Any, ...]]:
    """
    Returns a pair of nested Tuples containing a list of types and a list of input
    values sorted by the order specified by the ``abi``.

    ``normalized_args`` can be obtained by using
    :py:meth:`eth_utils.abi.get_normalized_abi_inputs`, which returns nested mappings
    or sequences corresponding to tuple-encoded values in ``abi``.

    :param abi_element: ABI element.
    :type abi_element: `ABIElement`
    :param normalized_args: Normalized arguments for the function.
    :type normalized_args: `Union[Tuple[Any, ...], Mapping[Any, Any]]`
    :return: Tuple of types and aligned arguments.
    :rtype: `Tuple[Tuple[str, ...], Tuple[Any, ...]]`

    .. doctest::

        >>> from eth_utils import get_aligned_abi_inputs
        >>> abi = {
        ...   'constant': False,
        ...   'inputs': [
        ...     {
        ...       'name': 'name',
        ...       'type': 'string'
        ...     },
        ...     {
        ...       'name': 's',
        ...       'type': 'uint256'
        ...     }
        ...   ],
        ...   'name': 'f',
        ...   'outputs': [],
        ...   'payable': False,
        ...   'stateMutability': 'nonpayable',
        ...   'type': 'function'
        ... }
        >>> get_aligned_abi_inputs(abi, ('myName', 123))
        (('string', 'uint256'), ('myName', 123))
    """
    _raise_if_fallback_or_receive_abi(abi_element)

    abi_element_inputs = cast(Sequence[ABIComponent], abi_element.get("inputs", []))
    if isinstance(normalized_args, abc.Mapping):
        # `args` is mapping.  Align values according to abi order.
        normalized_args = tuple(
            normalized_args[abi["name"]] for abi in abi_element_inputs
        )

    return (
        tuple(collapse_if_tuple(abi) for abi in abi_element_inputs),
        type(normalized_args)(
            _align_abi_input(abi, arg)
            for abi, arg in zip(abi_element_inputs, normalized_args)
        ),
    )


def get_abi_input_names(abi_element: ABIElement) -> List[Optional[str]]:
    """
    Return names for each input from the function or event ABI.

    :param abi_element: ABI element.
    :type abi_element: `ABIElement`
    :return: Names for each input in the function or event ABI.
    :rtype: `List[Optional[str]]`

    .. doctest::

        >>> from eth_utils import get_abi_input_names
        >>> abi = {
        ...   'constant': False,
        ...   'inputs': [
        ...     {
        ...       'name': 's',
        ...       'type': 'uint256'
        ...     }
        ...   ],
        ...   'name': 'f',
        ...   'outputs': [],
        ...   'payable': False,
        ...   'stateMutability': 'nonpayable',
        ...   'type': 'function'
        ... }
        >>> get_abi_input_names(abi)
        ['s']
    """
    _raise_if_fallback_or_receive_abi(abi_element)
    return [
        arg.get("name", None)
        for arg in cast(Sequence[ABIComponent], abi_element.get("inputs", []))
    ]


def get_abi_input_types(abi_element: ABIElement) -> List[str]:
    """
    Return types for each input from the function or event ABI.

    :param abi_element: ABI element.
    :type abi_element: `ABIElement`
    :return: Types for each input in the function or event ABI.
    :rtype: `List[str]`

    .. doctest::

        >>> from eth_utils import get_abi_input_types
        >>> abi = {
        ...   'constant': False,
        ...   'inputs': [
        ...     {
        ...       'name': 's',
        ...       'type': 'uint256'
        ...     }
        ...   ],
        ...   'name': 'f',
        ...   'outputs': [],
        ...   'payable': False,
        ...   'stateMutability': 'nonpayable',
        ...   'type': 'function'
        ... }
        >>> get_abi_input_types(abi)
        ['uint256']
    """
    _raise_if_fallback_or_receive_abi(abi_element)
    return [
        collapse_if_tuple(arg)
        for arg in cast(Sequence[ABIComponent], abi_element.get("inputs", []))
    ]


def get_abi_output_names(abi_element: ABIElement) -> List[Optional[str]]:
    """
    Return names for each output from the ABI element.

    :param abi_element: ABI element.
    :type abi_element: `ABIElement`
    :return: Names for each function output in the function ABI.
    :rtype: `List[Optional[str]]`

    .. doctest::

        >>> from eth_utils import get_abi_output_names
        >>> abi = {
        ...   'constant': False,
        ...   'inputs': [
        ...     {
        ...       'name': 's',
        ...       'type': 'uint256'
        ...     }
        ...   ],
        ...   'name': 'f',
        ...   'outputs': [
        ...     {
        ...       'name': 'name',
        ...       'type': 'string'
        ...     },
        ...     {
        ...       'name': 's',
        ...       'type': 'uint256'
        ...     }
        ...   ],
        ...   'payable': False,
        ...   'stateMutability': 'nonpayable',
        ...   'type': 'function'
        ... }
        >>> get_abi_output_names(abi)
        ['name', 's']
    """
    _raise_if_not_function_abi(abi_element)
    return [
        arg.get("name", None)
        for arg in cast(Sequence[ABIComponent], abi_element.get("outputs", []))
    ]


def get_abi_output_types(abi_element: ABIElement) -> List[str]:
    """
    Return types for each output from the function ABI.

    :param abi_element: ABI element.
    :type abi_element: `ABIElement`
    :return: Types for each function output in the function ABI.
    :rtype: `List[str]`

    .. doctest::

        >>> from eth_utils import get_abi_output_types
        >>> abi = {
        ...   'constant': False,
        ...   'inputs': [
        ...     {
        ...       'name': 's',
        ...       'type': 'uint256'
        ...     }
        ...   ],
        ...   'name': 'f',
        ...   'outputs': [
        ...     {
        ...       'name': 'name',
        ...       'type': 'string'
        ...     },
        ...     {
        ...       'name': 's',
        ...       'type': 'uint256'
        ...     }
        ...   ],
        ...   'payable': False,
        ...   'stateMutability': 'nonpayable',
        ...   'type': 'function'
        ... }
        >>> get_abi_output_types(abi)
        ['string', 'uint256']

    """
    _raise_if_not_function_abi(abi_element)
    return [
        collapse_if_tuple(arg)
        for arg in cast(Sequence[ABIComponent], abi_element.get("outputs", []))
    ]


def function_signature_to_4byte_selector(function_signature: str) -> bytes:
    r"""
    Return the 4-byte function selector from a function signature string.

    :param function_signature: String representation of the function name and arguments.
    :type function_signature: `str`
    :return: 4-byte function selector.
    :rtype: `bytes`

    .. doctest::

        >>> from eth_utils import function_signature_to_4byte_selector
        >>> function_signature_to_4byte_selector('myFunction()')
        b'\xc3x\n:'
    """
    return keccak(text=function_signature.replace(" ", ""))[:4]


def function_abi_to_4byte_selector(abi_element: ABIElement) -> bytes:
    r"""
    Return the 4-byte function signature of the provided function ABI.

    :param abi_element: ABI element.
    :type abi_element: `ABIElement`
    :return: 4-byte function signature.
    :rtype: `bytes`

    .. doctest::

        >>> from eth_utils import function_abi_to_4byte_selector
        >>> abi_element = {
        ...   'type': 'function',
        ...   'name': 'myFunction',
        ...   'inputs': [],
        ...   'outputs': []
        ... }
        >>> function_abi_to_4byte_selector(abi_element)
        b'\xc3x\n:'
    """
    function_signature = abi_to_signature(abi_element)
    return function_signature_to_4byte_selector(function_signature)


def event_signature_to_log_topic(event_signature: str) -> bytes:
    r"""
    Return the 32-byte keccak signature of the log topic for an event signature.

    :param event_signature: String representation of the event name and arguments.
    :type event_signature: `str`
    :return: Log topic bytes.
    :rtype: `bytes`

    .. doctest::

        >>> from eth_utils import event_signature_to_log_topic
        >>> event_signature_to_log_topic('MyEvent()')
        b'M\xbf\xb6\x8bC\xdd\xdf\xa1+Q\xeb\xe9\x9a\xb8\xfd\xedb\x0f\x9a\n\xc21B\x87\x9aO\x19*\x1byR\xd2'
    """
    return keccak(text=event_signature.replace(" ", ""))


def event_abi_to_log_topic(event_abi: ABIEvent) -> bytes:
    r"""
    Return the 32-byte keccak signature of the log topic from an event ABI.

    :param event_abi: Event ABI.
    :type event_abi: `ABIEvent`
    :return: Log topic bytes.
    :rtype: `bytes`

    .. doctest::

        >>> from eth_utils import event_abi_to_log_topic
        >>> abi = {
        ...   'type': 'event',
        ...   'anonymous': False,
        ...   'name': 'MyEvent',
        ...   'inputs': []
        ... }
        >>> event_abi_to_log_topic(abi)
        b'M\xbf\xb6\x8bC\xdd\xdf\xa1+Q\xeb\xe9\x9a\xb8\xfd\xedb\x0f\x9a\n\xc21B\x87\x9aO\x19*\x1byR\xd2'
    """
    event_signature = abi_to_signature(event_abi)
    return event_signature_to_log_topic(event_signature)



import pytest

from eth_utils.hexadecimal import (
    is_0x_prefixed,
)


@pytest.mark.parametrize(
    "value,expected",
    (("", False), ("0x", True), ("0x12345", True), ("12345", False), ("0X12345", True)),
)
def test_is_0x_prefixed(value, expected):
    assert is_0x_prefixed(value) is expected


@pytest.mark.parametrize("value", (b"", 123, {}, lambda: None))
def test_is_0x_prefixed_rejects_non_text_types(value):
    with pytest.raises(TypeError):
        is_0x_prefixed(value)



from pydantic import (
    Field,
)

from eth_utils.pydantic import (
    CamelModel,
)


class PydanticTestClass(CamelModel):
    field_five: int = 5
    field_six: str = "6"
    field_seven: str = Field(default="7", exclude=True)
    field_eight: int = Field(default=8, exclude=True)


class NestedPydanticTestClass(CamelModel):
    field_one: int = 1
    field_two: str = "2"
    field_three: str = Field(default="3", exclude=True)
    field_four: PydanticTestClass = PydanticTestClass()


def test_camel_model_dump():
    dump_alias = PydanticTestClass().model_dump(by_alias=True)
    dump = PydanticTestClass().model_dump()
    assert dump_alias == {"fieldFive": 5, "fieldSix": "6"}
    assert dump == {"field_five": 5, "field_six": "6"}

    dump_alias_nested = NestedPydanticTestClass().model_dump(by_alias=True)
    dump_nested = NestedPydanticTestClass().model_dump()
    assert dump_alias_nested == {
        "fieldOne": 1,
        "fieldTwo": "2",
        "fieldFour": {"fieldFive": 5, "fieldSix": "6"},
    }
    assert dump_nested == {
        "field_one": 1,
        "field_two": "2",
        "field_four": {"field_five": 5, "field_six": "6"},
    }


def test_camel_model_json_schema():
    schema = PydanticTestClass().model_json_schema(by_alias=True)
    assert isinstance(schema, dict)
    properties = schema["properties"]
    assert len(properties) == 2
    assert all(field in properties for field in {"fieldFive", "fieldSix"})
    assert "fieldSix" in properties
    field_5_props = {"title": "Fieldfive", "type": "integer", "default": 5}
    assert properties["fieldFive"] == field_5_props
    field_6_props = {"title": "Fieldsix", "type": "string", "default": "6"}
    assert properties["fieldSix"] == field_6_props

    schema_nested = NestedPydanticTestClass().model_json_schema(by_alias=True)
    assert isinstance(schema_nested, dict)
    properties_nested = schema_nested["properties"]
    assert len(properties_nested) == 3
    assert all(
        field in properties_nested for field in {"fieldOne", "fieldTwo", "fieldFour"}
    )
    assert properties_nested["fieldOne"] == {
        "title": "Fieldone",
        "type": "integer",
        "default": 1,
    }
    assert properties_nested["fieldTwo"] == {
        "title": "Fieldtwo",
        "type": "string",
        "default": "2",
    }
    assert properties_nested["fieldFour"] == {
        "$ref": "#/$defs/PydanticTestClass",
        "default": {"fieldFive": 5, "fieldSix": "6"},
    }
    # assert the reference to the nested model properties
    assert schema_nested["$defs"]["PydanticTestClass"]["properties"] == properties



from typing import (
    List,
)

from eth_utils import (
    apply_to_return_value,
)


def wrap_as_list(value: int) -> List[int]:
    return [value]


@apply_to_return_value(wrap_as_list)
def return_value() -> int:
    return 1


x = return_value()
reveal_type(x)  # noqa: F821



import pytest

from hypothesis import (
    given,
    strategies as st,
)

from eth_utils.encoding import (
    big_endian_to_int,
    int_to_big_endian,
)


@pytest.mark.parametrize(
    "as_int,as_big_endian",
    (
        (0, b"\x00"),
        (1, b"\x01"),
        (7, b"\x07"),
        (8, b"\x08"),
        (9, b"\x09"),
        (256, b"\x01\x00"),
        (2**256 - 1, b"\xff" * 32),
    ),
)
def test_big_endian_conversions(as_int, as_big_endian):
    as_int_result = big_endian_to_int(as_big_endian)
    assert as_int_result == as_int

    as_big_endian_result = int_to_big_endian(as_int)
    assert as_big_endian_result == as_big_endian


@given(value=st.integers(min_value=0, max_value=2**256 - 1))
def test_big_endian_round_trip_from_int(value):
    result = big_endian_to_int(int_to_big_endian(value))
    assert result == value


@given(
    value=st.binary(min_size=1, max_size=32).map(lambda v: v.lstrip(b"\x00") or b"\x00")
)
def test_big_endian_round_trip_from_big_endian(value):
    result = int_to_big_endian(big_endian_to_int(value))
    assert result == value



import collections

from eth_utils import (
    flatten_return,
    reversed_return,
    sort_return,
    to_dict,
    to_list,
    to_ordered_dict,
    to_set,
    to_tuple,
)


def yield_things(*things):
    yield from things


def test_to_tuple():
    fn = to_tuple(yield_things)
    actual = fn(1, 1, 2, 3, 5, 8)
    assert actual == (1, 1, 2, 3, 5, 8)


def test_to_list():
    fn = to_list(yield_things)
    actual = fn(1, 1, 2, 3, 5, 8)
    assert actual == [1, 1, 2, 3, 5, 8]


def test_to_dict():
    fn = to_dict(yield_things)
    actual = fn(("a", 1), ("b", 2), ("c", 3))
    assert actual == {"a": 1, "b": 2, "c": 3}


def test_to_ordered_dict():
    fn = to_ordered_dict(yield_things)
    actual = fn(("b", 2), ("d", 4), ("a", 1), ("c", 3))
    assert actual == collections.OrderedDict((("b", 2), ("d", 4), ("a", 1), ("c", 3)))
    assert tuple(actual.items()) == (("b", 2), ("d", 4), ("a", 1), ("c", 3))


def test_to_set():
    fn = to_set(yield_things)
    actual = fn("a", "b", "a", "c")
    assert actual == {"a", "b", "c"}


def test_sorted_return():
    fn = sort_return(yield_things)
    actual = fn(1, 8, 2, 4, 5, 9, 3)
    assert actual == (1, 2, 3, 4, 5, 8, 9)


def test_flatten_return():
    fn = flatten_return(yield_things)
    actual = fn((1, 3, 4), (2, 1), (8, 5))
    assert actual == (1, 3, 4, 2, 1, 8, 5)


def test_reversed_return():
    fn = reversed_return(yield_things)
    actual = fn(1, 5, 2, 4, 3)
    assert actual == (3, 4, 2, 5, 1)



import pytest

from eth_utils import (
    replace_exceptions,
)


@pytest.fixture()
def mock_function_with_exception(old_to_new):
    @replace_exceptions(old_to_new)
    def function_with_exception(x):
        raise TypeError("Boom!")

    return function_with_exception


@pytest.mark.parametrize(
    "old_to_new,new",
    (
        ({TypeError: AttributeError}, AttributeError),
        ({TypeError: NameError}, NameError),
        ({ValueError: AttributeError, TypeError: NameError}, NameError),
    ),
)
def test_decorator_replaces_exceptions(mock_function_with_exception, old_to_new, new):
    with pytest.raises(new, match="Boom!"):
        mock_function_with_exception(old_to_new)



import pytest

from eth_utils import (
    denoms,
    humanize_wei,
)


@pytest.mark.parametrize(
    "value,expected",
    (
        (1, "1 wei"),
        (denoms.wei, "1 wei"),
        (denoms.kwei, "1000 wei"),
        (denoms.babbage, "1000 wei"),
        (denoms.femtoether, "1000 wei"),
        (denoms.mwei, "0.001 gwei"),
        (denoms.lovelace, "0.001 gwei"),
        (denoms.picoether, "0.001 gwei"),
        (denoms.gwei, "1 gwei"),
        (denoms.shannon, "1 gwei"),
        (denoms.nanoether, "1 gwei"),
        (denoms.nano, "1 gwei"),
        (denoms.szabo, "1000 gwei"),
        (denoms.microether, "1000 gwei"),
        (denoms.micro, "1000 gwei"),
        (denoms.finney, "0.001 ether"),
        (denoms.milliether, "0.001 ether"),
        (denoms.milli, "0.001 ether"),
        (denoms.ether, "1 ether"),
        (denoms.kether, "1000 ether"),
        (denoms.grand, "1000 ether"),
        (denoms.mether, "1000000 ether"),
        (denoms.gether, "1000000000 ether"),
        (denoms.tether, "1000000000000 ether"),
        (891, "891 wei"),
        (987654, "987654 wei"),
        (9876543, "0.009876543 gwei"),
        (100 * denoms.finney, "0.1 ether"),
        (denoms.finney - 1, "999999.999999999 gwei"),
        (denoms.finney + 1, "0.001000000000000001 ether"),
        (0, "0 wei"),
    ),
)
def test_humanize_wei(value, expected):
    assert humanize_wei(value) == expected


def test_humanize_error_signed_int():
    with pytest.raises(ValueError):
        humanize_wei(-1)


def test_humanize_error_string():
    with pytest.raises(TypeError):
        humanize_wei("hey")



"""
Script that syncs networks JSON with the current list of networks on chainid.network

Update the networks_file by running this script:

.. code-block:: shell

    $ python update_networks.py

After running the command, check the output in the networks_file.
If all looks good, open a PR with the changes.
"""

import urllib.request

networks_file = "./eth_utils/__json/eth_networks.json"

with urllib.request.urlopen("https://chainid.network/chains_mini.json") as response:
    content = response.read().decode()

with open(networks_file, "w") as open_file:
    open_file.write(f"{content}\n")

print("Networks have updated! Please review and open a PR with the changes.")



import pytest
import collections

from pydantic import (
    Field,
)

import eth_utils
from eth_utils.curried import (
    apply_formatter_at_index,
    apply_formatter_if,
    apply_formatter_to_array,
    apply_formatters_to_dict,
    apply_formatters_to_sequence,
    apply_key_map,
    apply_one_of_formatters,
    is_list_like,
    is_string,
)
from eth_utils.pydantic import (
    CamelModel,
)


def i_put_my_thing_down_flip_it_and_reverse_it(lyric):
    return "".join(reversed(lyric))


CONDITION_FORMATTER_PAIRS = (
    (is_list_like, tuple),
    (is_string, i_put_my_thing_down_flip_it_and_reverse_it),
)


def test_format_dict_error():
    with pytest.raises(ValueError) as exc_info:
        apply_formatters_to_dict({"myfield": int}, {"myfield": "a"})
    with pytest.raises(ValueError) as exc_info:
        eth_utils.apply_formatters_to_dict({"myfield": int}, {"myfield": "a"})
    assert "myfield" in str(exc_info.value)


class PydanticModel(CamelModel):
    to_bytes_from_int: int = 1
    to_int_from_bytes: bytes = b"\x02"
    excluded_field: int = Field(default=3, exclude=True)
    excluded_field_two: str = Field(default="4", exclude=True)


@pytest.mark.parametrize(
    "formatter,value,expected,unaliased",
    (
        (
            {"should_be_int": int, "should_be_bool": bool},
            {"should_be_int": 1.2, "should_be_bool": 3.4, "pass_through": 5.6},
            {"should_be_int": 1, "should_be_bool": True, "pass_through": 5.6},
            False,  # does not apply
        ),
        (
            {"toBytesFromInt": eth_utils.to_bytes, "toIntFromBytes": eth_utils.to_int},
            PydanticModel(),
            {"toBytesFromInt": b"\x01", "toIntFromBytes": 2},
            False,
        ),
        (
            {
                "to_bytes_from_int": eth_utils.to_bytes,
                "to_int_from_bytes": eth_utils.to_int,
            },
            PydanticModel(),
            {"to_bytes_from_int": b"\x01", "to_int_from_bytes": 2},
            True,
        ),
    ),
)
def test_apply_formatters_to_dict(formatter, value, expected, unaliased):
    assert (
        eth_utils.apply_formatters_to_dict(formatter, value, unaliased=unaliased)
        == expected
    )

    mapper = apply_formatters_to_dict(formatter, unaliased=unaliased)
    assert mapper(value) == expected


@pytest.mark.parametrize(
    "formatter, value, expected",
    (
        (
            {"black": "orange", "Internet": "Ethereum"},
            {"black": 1.2, "Internet": 3.4, "pass_through": 5.6},
            {"orange": 1.2, "Ethereum": 3.4, "pass_through": 5.6},
        ),
    ),
)
def test_apply_key_map(formatter, value, expected):
    assert eth_utils.apply_key_map(formatter, value) == expected

    mapper = apply_key_map(formatter)
    assert mapper(value) == expected


@pytest.mark.parametrize(
    "formatter, value",
    (
        ({"a": "b"}, {"b": 3}),
        ({"a": "b"}, {"a": 2, "b": 3}),
        ({"a": "b"}, collections.OrderedDict((("a", 2), ("b", 3)))),
        ({"a": "b"}, collections.OrderedDict((("b", 3), ("a", 2)))),
    ),
)
def test_apply_key_map_with_key_conflicts_raises_exception(formatter, value):
    with pytest.raises(KeyError):
        eth_utils.apply_key_map(formatter, value)


@pytest.mark.parametrize(
    "condition, formatter, value, expected",
    (
        (is_string, bool, 1, 1),
        (is_string, bool, "1", True),
        (is_string, bool, "", False),
    ),
)
def test_apply_formatter_if(condition, formatter, value, expected):
    assert eth_utils.apply_formatter_if(condition, formatter, value) == expected

    # must be able to curry
    conditional_formatter = apply_formatter_if(condition, formatter)
    assert conditional_formatter(value) == expected


@pytest.mark.parametrize(
    "condition_formatters, value, expected",
    (
        (CONDITION_FORMATTER_PAIRS, "my thing", "gniht ym"),
        (CONDITION_FORMATTER_PAIRS, [2, 3], (2, 3)),
        (CONDITION_FORMATTER_PAIRS, 1, ValueError),
    ),
)
def test_apply_one_of_formatters(condition_formatters, value, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            apply_one_of_formatters(condition_formatters, value)
        with pytest.raises(expected):
            eth_utils.apply_one_of_formatters(condition_formatters, value)
    else:
        assert (
            eth_utils.apply_one_of_formatters(condition_formatters, value) == expected
        )

        # must be able to curry
        apply_one = apply_one_of_formatters(condition_formatters)
        assert apply_one(value) == expected


@pytest.mark.parametrize(
    "formatter, index, value, expected",
    (
        (bool, 1, [1, 2, 3], [1, True, 3]),
        (bool, 1, (1, 2, 3), (1, True, 3)),
        (bool, 3, (1, 2, 3), IndexError),
    ),
)
def test_apply_formatter_at_index(formatter, index, value, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            apply_formatter_at_index(formatter, index, value)
        with pytest.raises(expected):
            eth_utils.apply_formatter_at_index(formatter, index, value)
    else:
        assert eth_utils.apply_formatter_at_index(formatter, index, value) == expected

        # must be able to curry
        targetted_formatter = apply_formatter_at_index(formatter, index)
        assert targetted_formatter(value) == expected


SEQUENCE_FORMATTER_PARAMETERS = (
    ([bool, int, str], (1.2, 3.4, 5.6), (True, 3, "5.6")),
    ([bool, int, str], [1.2, 3.4, 5.6], [True, 3, "5.6"]),
    ([bool, int, str, float], (1.2, 3.4, 5.6), IndexError),
)

LOOSE_SEQUENCE_FORMATTER_PARAMETERS = SEQUENCE_FORMATTER_PARAMETERS + (
    ([bool, int], (1.2, 3.4, 5.6), (True, 3, 5.6)),
)


@pytest.mark.parametrize(
    "formatters, value, expected", LOOSE_SEQUENCE_FORMATTER_PARAMETERS
)
def test_combine_argument_formatters(formatters, value, expected):
    with pytest.warns(DeprecationWarning):
        list_formatter = eth_utils.combine_argument_formatters(*formatters)
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            list_formatter(value)
    else:
        assert list_formatter(value) == expected


STRICT_SEQUENCE_FORMATTER_PARAMETERS = SEQUENCE_FORMATTER_PARAMETERS + (
    ([bool, int], (1.2, 3.4, 5.6), IndexError),
)


@pytest.mark.parametrize(
    "formatters, value, expected", STRICT_SEQUENCE_FORMATTER_PARAMETERS
)
def test_apply_formatters_to_sequence_curried(formatters, value, expected):
    list_formatter = apply_formatters_to_sequence(formatters)
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            list_formatter(value)
    else:
        assert list_formatter(value) == expected


@pytest.mark.parametrize(
    "formatters, value, expected", STRICT_SEQUENCE_FORMATTER_PARAMETERS
)
def test_apply_formatters_to_sequence(formatters, value, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            eth_utils.apply_formatters_to_sequence(formatters, value)
    else:
        assert eth_utils.apply_formatters_to_sequence(formatters, value) == expected


@pytest.mark.parametrize(
    "formatter, value, expected",
    ((int, [1.2, 3.4, 5.6], [1, 3, 5]), (int, (1.2, 3.4, 5.6), (1, 3, 5))),
)
def test_apply_formatter_to_array(formatter, value, expected):
    assert eth_utils.apply_formatter_to_array(formatter, value) == expected

    mapper = apply_formatter_to_array(formatter)
    assert mapper(value) == expected



# no-unused-vars
from typing import (
    Sequence,
)

from eth_utils import (
    filter_abi_by_type,
)
from eth_utils.abi import (
    ABIConstructor,
    ABIError,
    ABIEvent,
    ABIFallback,
    ABIFunction,
    ABIReceive,
)


def test_filter_abi_by_type_function() -> None:
    _result: Sequence[ABIFunction] = filter_abi_by_type("function", [])  # noqa: F841


def test_filter_abi_by_type_constructor() -> None:
    _result: Sequence[ABIConstructor] = filter_abi_by_type(  # noqa: F841
        "constructor", []
    )


def test_filter_abi_by_type_fallback() -> None:
    _result: Sequence[ABIFallback] = filter_abi_by_type("fallback", [])  # noqa: F841


def test_filter_abi_by_type_receive() -> None:
    _result: Sequence[ABIReceive] = filter_abi_by_type("receive", [])  # noqa: F841


def test_filter_abi_by_type_event() -> None:
    _result: Sequence[ABIEvent] = filter_abi_by_type("event", [])  # noqa: F841


def test_filter_abi_by_type_error() -> None:
    _result: Sequence[ABIError] = filter_abi_by_type("error", [])  # noqa: F841



import pytest

from eth_utils import (
    humanize_hash,
)


@pytest.mark.parametrize("hash32,expected", ((bytes(range(32)), "0001..1e1f"),))
def test_humanize_hash(hash32, expected):
    assert humanize_hash(hash32) == expected



import pytest

from eth_utils.decorators import (
    validate_conversion_arguments,
)


@pytest.fixture()
def mock_conversion_function():
    @validate_conversion_arguments
    def conversion_function(primitive=None, hexstr=None, text=None):
        return True

    return conversion_function


@pytest.mark.parametrize("val", (b"123", 123, {}, lambda: None))
def test_decorator_rejects_non_text_type_args_for_hexstr_kwarg(
    mock_conversion_function, val
):
    with pytest.raises(TypeError):
        mock_conversion_function(hexstr=val)


@pytest.mark.parametrize("val", (b"123", 123, {}, lambda: None))
def test_decorator_rejects_non_text_type_args_for_text_kwarg(
    mock_conversion_function, val
):
    with pytest.raises(TypeError):
        mock_conversion_function(text=val)


@pytest.mark.parametrize(
    "values",
    (
        {},
        {"primitive": "abc", "hexstr": "abc"},
        {"primitive": "abc", "hexstr": "abc", "text": "abc"},
    ),
)
def test_decorator_only_accepts_a_single_arg(mock_conversion_function, values):
    with pytest.raises(TypeError):
        mock_conversion_function(**values)


def test_decorator_rejects_invalid_kwarg(mock_conversion_function):
    with pytest.raises(TypeError):
        mock_conversion_function(value="123")


@pytest.mark.parametrize(
    "values", ({"primitive": "abc"}, {"hexstr": "abc"}, {"text": "abc"})
)
def test_decorator_accepts_a_single_text_arg(mock_conversion_function, values):
    actual = mock_conversion_function(**values)
    assert actual is True



import pytest
import logging
import uuid

from eth_utils import (
    ExtendedDebugLogger,
)
from eth_utils.logging import (
    DEBUG2_LEVEL_NUM,
)


@pytest.fixture
def DEBUG2_installed():
    # caplog only works on loggers retrieved from `logging.getLogger` so we
    # have to use that mechanism to instantiate our logger.
    original_logger = logging.getLoggerClass()
    logging.setLoggerClass(ExtendedDebugLogger)

    try:
        yield
    finally:
        logging.setLoggerClass(original_logger)


def test_extended_debug_logger(caplog, DEBUG2_installed):
    caplog.set_level(DEBUG2_LEVEL_NUM, "testing")

    logger = logging.getLogger("testing")

    assert len(caplog.records) == 0
    logger.debug2("message %s", 1)
    assert len(caplog.records) == 1

    record = caplog.records[0]
    assert record.levelno == 8
    assert record.args == (1,)
    assert record.message == "message 1"


def test_caching_of_debug2_when_disabled(caplog, DEBUG2_installed):
    caplog.set_level(DEBUG2_LEVEL_NUM, "testing")

    # we need a unique logger because loggers are cached
    logger = logging.getLogger(f"testing-{uuid.uuid4()}")

    assert logger.isEnabledFor(DEBUG2_LEVEL_NUM) is False

    assert "show_debug2" not in logger.__dict__
    assert logger.show_debug2 is False
    # cached property should have inserted it into the dict
    assert "show_debug2" in logger.__dict__

    # sanity pre-check
    assert len(caplog.records) == 0

    assert "debug2" not in logger.__dict__
    assert logger.debug2("this should actually call the function") is None
    assert "debug2" in logger.__dict__
    assert logger.debug2("should not do anything but hit the lambda") is None

    assert len(caplog.records) == 0

    # now see that it always returns the value from `__dict__`
    logger.__dict__["show_debug2"] = 100
    assert logger.show_debug2 == 100


def test_caching_of_debug2_when_enabled(caplog, DEBUG2_installed):
    caplog.set_level(DEBUG2_LEVEL_NUM, "testing")

    # we need a unique logger because loggers are cached
    logger = logging.getLogger(f"testing-{uuid.uuid4()}")

    assert logger.isEnabledFor(DEBUG2_LEVEL_NUM) is False
    logger.setLevel(DEBUG2_LEVEL_NUM)
    assert logger.isEnabledFor(DEBUG2_LEVEL_NUM) is True

    assert "show_debug2" not in logger.__dict__
    assert logger.show_debug2 is True
    # cached property should have inserted it into the dict
    assert "show_debug2" in logger.__dict__

    # sanity pre-check
    assert len(caplog.records) == 0

    assert "debug2" not in logger.__dict__
    assert logger.debug2("this should actually call the function") is None
    assert "debug2" not in logger.__dict__
    assert logger.debug2("this should still call the function") is None

    assert len(caplog.records) == 2

    # now see that it always returns the value from `__dict__`
    logger.__dict__["show_debug2"] = 100
    assert logger.show_debug2 == 100



#!/usr/bin/env python3

import os
import sys
from pathlib import Path
import subprocess


def main():
    template_dir = Path(os.path.dirname(sys.argv[0]))
    template_vars_file = template_dir / "template_vars.txt"
    fill_template_vars_script = template_dir / "fill_template_vars.py"

    with open(template_vars_file, "r") as input_file:
        content_lines = input_file.readlines()

    process = subprocess.Popen(
        [sys.executable, str(fill_template_vars_script)],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    for line in content_lines:
        process.stdin.write(line)
        process.stdin.flush()

    stdout, stderr = process.communicate()

    if process.returncode != 0:
        print(f"Error occurred: {stderr}")
        sys.exit(1)

    print(stdout)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Towncrier silently ignores files that do not match the expected ending.
# We use this script to ensure we catch these as errors in CI.

import pathlib
import sys

ALLOWED_EXTENSIONS = {
    ".breaking.rst",
    ".bugfix.rst",
    ".deprecation.rst",
    ".docs.rst",
    ".feature.rst",
    ".internal.rst",
    ".misc.rst",
    ".performance.rst",
    ".removal.rst",
}

ALLOWED_FILES = {
    "validate_files.py",
    "README.md",
}

THIS_DIR = pathlib.Path(__file__).parent

num_args = len(sys.argv) - 1
assert num_args in {0, 1}
if num_args == 1:
    assert sys.argv[1] in ("is-empty",)

for fragment_file in THIS_DIR.iterdir():
    if fragment_file.name in ALLOWED_FILES:
        continue
    elif num_args == 0:
        full_extension = "".join(fragment_file.suffixes)
        if full_extension not in ALLOWED_EXTENSIONS:
            raise Exception(f"Unexpected file: {fragment_file}")
    elif sys.argv[1] == "is-empty":
        raise Exception(f"Unexpected file: {fragment_file}")
    else:
        raise RuntimeError(
            f"Strange: arguments {sys.argv} were validated, but not found"
        )



try:
    from cytoolz import (
        curry,
        keyfilter,
        merge_with,
        valfilter,
    )
    from cytoolz.functoolz import (
        Compose,
        has_keywords,
        num_required_args,
    )
except ImportError:
    from toolz import curry, keyfilter, merge_with, valfilter
    from toolz.functoolz import Compose, has_keywords, num_required_args

import eth_utils
import eth_utils.curried


def enhanced_num_required_args(func):
    if isinstance(func, Compose):
        return enhanced_num_required_args(func.first)
    else:
        return num_required_args(func)


def enhanced_has_keywords(func):
    if isinstance(func, Compose):
        return enhanced_has_keywords(func.first)
    else:
        return has_keywords(func)


# heavily inspired by https://github.com/pytoolz/toolz/blob/20d8aefc0a5/toolz/tests/test_curried.py  # noqa: E501
def test_curried_namespace():
    def should_curry(value):
        if not callable(value) or isinstance(value, curry) or isinstance(value, type):
            return False
        if isinstance(value, type) and issubclass(value, Exception):
            return False
        nargs = enhanced_num_required_args(value)
        if nargs is None or nargs > 1:
            return True
        else:
            return nargs == 1 and enhanced_has_keywords(value)

    def curry_namespace(ns):
        return {
            name: curry(f) if should_curry(f) else f
            for name, f in ns.items()
            if "__" not in name
        }

    all_auto_curried = curry_namespace(vars(eth_utils))

    inferred_namespace = valfilter(callable, all_auto_curried)
    curried_namespace = valfilter(callable, eth_utils.curried.__dict__)

    if inferred_namespace != curried_namespace:
        missing = set(inferred_namespace) - set(curried_namespace)
        if missing:
            to_insert = sorted(f"{f}," for f in missing)
            raise AssertionError(
                "There are missing functions in eth_utils.curried:\n"
                + "\n".join(to_insert)
            )
        extra = set(curried_namespace) - set(inferred_namespace)
        if extra:
            raise AssertionError(
                "There are extra functions in eth_utils.curried:\n"
                + "\n".join(sorted(extra))
            )
        unequal = merge_with(list, inferred_namespace, curried_namespace)
        unequal = valfilter(lambda x: x[0] != x[1], unequal)
        to_curry = keyfilter(lambda x: should_curry(getattr(eth_utils, x)), unequal)
        if to_curry:
            to_curry_formatted = sorted(f"{f} = curry({f})" for f in to_curry)
            raise AssertionError(
                "There are missing functions to curry in eth_utils.curried:\n"
                + "\n".join(to_curry_formatted)
            )
        elif unequal:
            not_to_curry_formatted = sorted(unequal)
            raise AssertionError(
                "Missing functions NOT to curry in eth_utils.curried:\n"
                + "\n".join(not_to_curry_formatted)
            )
        else:
            raise AssertionError(
                f"unexplained difference between {repr(inferred_namespace)} and "
                f"{repr(curried_namespace)}"
            )



import pytest

from eth_utils import (
    humanize_bytes,
)


@pytest.mark.parametrize(
    "value,expected",
    (
        (bytes(range(32)), "0001..1e1f"),
        (bytes(range(3)), "000102"),
        (bytes(range(5)), "0001020304"),
        (bytes(range(6)), "0001..0405"),
    ),
)
def test_humanize_bytes(value, expected):
    assert humanize_bytes(value) == expected



#!/usr/bin/env python3

import os
import sys
import re
from pathlib import Path


def _find_files(project_root):
    path_exclude_pattern = r"\.git($|\/)|venv|_build"
    file_exclude_pattern = r"fill_template_vars\.py|\.swp$"
    filepaths = []
    for dir_path, _dir_names, file_names in os.walk(project_root):
        if not re.search(path_exclude_pattern, dir_path):
            for file in file_names:
                if not re.search(file_exclude_pattern, file):
                    filepaths.append(str(Path(dir_path, file)))

    return filepaths


def _replace(pattern, replacement, project_root):
    print(f"Replacing values: {pattern}")
    for file in _find_files(project_root):
        try:
            with open(file) as f:
                content = f.read()
            content = re.sub(pattern, replacement, content)
            with open(file, "w") as f:
                f.write(content)
        except UnicodeDecodeError:
            pass


def main():
    project_root = Path(os.path.realpath(sys.argv[0])).parent.parent

    module_name = input("What is your python module name? ")

    pypi_input = input(f"What is your pypi package name? (default: {module_name}) ")
    pypi_name = pypi_input or module_name

    repo_input = input(f"What is your github project name? (default: {pypi_name}) ")
    repo_name = repo_input or pypi_name

    rtd_input = input(
        f"What is your readthedocs.org project name? (default: {pypi_name}) "
    )
    rtd_name = rtd_input or pypi_name

    project_input = input(
        f"What is your project name (ex: at the top of the README)? (default: {repo_name}) "
    )
    project_name = project_input or repo_name

    short_description = input("What is a one-liner describing the project? ")

    _replace("<MODULE_NAME>", module_name, project_root)
    _replace("<PYPI_NAME>", pypi_name, project_root)
    _replace("<REPO_NAME>", repo_name, project_root)
    _replace("<RTD_NAME>", rtd_name, project_root)
    _replace("<PROJECT_NAME>", project_name, project_root)
    _replace("<SHORT_DESCRIPTION>", short_description, project_root)

    os.makedirs(project_root / module_name, exist_ok=True)
    Path(project_root / module_name / "__init__.py").touch()
    Path(project_root / module_name / "py.typed").touch()


if __name__ == "__main__":
    main()



import logging
from typing import (
    Generic,
    TypeVar,
)

from eth_utils import (
    ExtendedDebugLogger,
    HasExtendedDebugLogger,
    HasLogger,
    HasLoggerMeta,
)


def verify_has_logger_type() -> logging.Logger:
    return HasLogger.logger


def verify_has_extended_debug_logger_type() -> ExtendedDebugLogger:
    return HasExtendedDebugLogger.logger


GenericWithLoggerMeta = HasLoggerMeta.meta_compat(type)


# mypy doesn't recognize this as a valid metaclass but it still functions
# properly
class GenericWithLogger(Generic, metaclass=GenericWithLoggerMeta):  # type: ignore
    pass


T = TypeVar("T")


# mypy gets confused about the actual provision of the types for the generics
# but the actual type checking mechanism below works.
class VerifyGenericWithLoggerTyping(GenericWithLogger[T]):  # type: ignore
    @classmethod
    def return_T(cls, value: T) -> T:
        return value


class VerifyGenericWithLoggerTypingInteger(VerifyGenericWithLoggerTyping[int]):
    pass


int_value: int

int_value = VerifyGenericWithLoggerTypingInteger.return_T(1234)

# mypy should be angry about passing a non-integer into this function. We test
# this by using an ignore comment relying on our type checking settings to warn
# on un-used ignore statements
VerifyGenericWithLoggerTypingInteger.return_T("not-an-integer")  # type: ignore



from typing import (
    Optional,
    Union,
)

from eth_hash.auto import (
    keccak as keccak_256,
)

from .conversions import (
    to_bytes,
)


def keccak(
    primitive: Optional[Union[bytes, int, bool]] = None,
    hexstr: Optional[str] = None,
    text: Optional[str] = None,
) -> bytes:
    return bytes(keccak_256(to_bytes(primitive, hexstr, text)))



import decimal
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    List,
    NoReturn,
    Set,
    Tuple,
)

from eth_utils import (
    clamp,
    denoms,
    replace_exceptions,
    to_dict,
    to_list,
    to_ordered_dict,
    to_set,
    to_tuple,
)

if TYPE_CHECKING:
    from collections import OrderedDict  # noqa: F401


@to_tuple
def typing_to_tuple() -> Iterable[int]:
    yield 1
    yield 2
    yield 3


v_tuple: Tuple[int, ...] = typing_to_tuple()


@to_list
def typing_to_list() -> Iterable[int]:
    yield 1
    yield 2
    yield 3


v_list: List[int] = typing_to_list()


@to_set
def typing_to_set() -> Iterable[int]:
    yield 1
    yield 2
    yield 3


v_set: Set[int] = typing_to_set()


@to_dict
def typing_to_dict() -> Iterable[Tuple[str, int]]:
    yield ("a", 1)
    yield ("b", 2)
    yield ("c", 3)


v_dict: Dict[str, int] = typing_to_dict()


@to_ordered_dict
def typing_to_ordered_dict() -> Iterable[Tuple[str, int]]:
    yield ("a", 1)
    yield ("b", 2)
    yield ("c", 3)


v_ordered_dict: "OrderedDict[str, int]" = typing_to_ordered_dict()

# verifies that the denoms object is properly typed.
ether: int = denoms.ether


@replace_exceptions({ValueError: TypeError})
def example_replace_exceptions() -> NoReturn:
    raise ValueError("The base exception")


def return_value_int(lower: int, upper: int, value: int) -> int:
    return clamp(lower, upper, value)


int_value: int = return_value_int(2, 5, 8)


def return_value_float(lower: float, upper: float, value: float) -> float:
    return clamp(lower, upper, value)


float_value: float = return_value_float(2.0, 5.0, 8.0)


def return_value_decimal(
    lower: decimal.Decimal, upper: decimal.Decimal, value: decimal.Decimal
) -> decimal.Decimal:
    return clamp(lower, upper, value)


decimal_value: decimal.Decimal = return_value_decimal(
    decimal.Decimal("2.0"), decimal.Decimal("5.0"), decimal.Decimal("8.0")
)



import pytest

from eth_utils.humanize import (
    humanize_ipfs_uri,
    is_ipfs_uri,
)


@pytest.mark.parametrize(
    "uri,expected",
    (
        ("ipfs://QmTKB75Y73zhNbD3Y73xeXGjYrZHmaXXNxoZqGCagu7r8u", "ipfs://QmTK..7r8u"),
        ("ipfs://QmaRBSoxprXdaEU98QqgtVkKrX4jcVxYdSR2TQwBKNiGMH", "ipfs://QmaR..iGMH"),
    ),
)
def test_humanize_ipfs_uri(uri, expected):
    actual = humanize_ipfs_uri(uri)
    assert actual == expected


@pytest.mark.parametrize(
    "uri,expected",
    (
        ("ipfs://QmTKB75Y73zhNbD3Y73xeXGjYrZHmaXXNxoZqGCagu7r8u", True),
        ("ipfs://QmaRBSoxprXdaEU98QqgtVkKrX4jcVxYdSR2TQwBKNiGMH", True),
        (123, False),
        ([], False),
        (False, False),
        ("ipfs://", False),
        ("ipfs//QmTKB75Y73zhNbD3Y73xeXGjYrZHmaXXNxoZqGCagu7r8u", False),
        ("ipfs:/QmTKB75Y73zhNbD3Y73xeXGjYrZHmaXXNxoZqGCagu7r8u", False),
        ("ipfs/QmTKB75Y73zhNbD3Y73xeXGjYrZHmaXXNxoZqGCagu7r8u", False),
        ("ipfsQmTKB75Y73zhNbD3Y73xeXGjYrZHmaXXNxoZqGCagu7r8u", False),
        ("http://QmTKB75Y73zhNbD3Y73xeXGjYrZHmaXXNxoZqGCagu7r8u", False),
    ),
)
def test_is_ipfs_uri(uri, expected):
    actual = is_ipfs_uri(uri)
    assert actual == expected



import pytest
from typing import (
    List,
)

from mypy import (
    api,
)

MYPY_ARGS = ["--ignore-missing-imports"]
FIXTURE_DIR = "fixtures/mypy/"


def fixture_dir(val: str):
    return FIXTURE_DIR + val


def check_mypy_run(
    cmd_line: List[str],
    expected_out: str,
    expected_err: str = "",
    expected_returncode: int = 0,
) -> None:
    """Helper to run mypy and check the output."""
    out, err, returncode = api.run(cmd_line)
    assert out == expected_out, err
    assert err == expected_err, out
    assert returncode == expected_returncode, returncode


# The following tests all run code snippets through mypy that contain a
# `reveal_type` statement, and then match the mypy output against our expectations
@pytest.mark.parametrize(
    "fixture,message",
    (
        (
            fixture_dir("to_tuple_decorator.py"),
            fixture_dir(
                'to_tuple_decorator.py:16: note: Revealed type is "builtins.tuple[builtins.int, ...]"\nSuccess: no issues found in 1 source file\n'  # noqa: E501
            ),
        ),
        (
            fixture_dir("to_list_decorator.py"),
            fixture_dir(
                'to_list_decorator.py:16: note: Revealed type is "builtins.list[builtins.int]"\nSuccess: no issues found in 1 source file\n'  # noqa: E501
            ),
        ),
        (
            fixture_dir("to_set_decorator.py"),
            fixture_dir(
                'to_set_decorator.py:16: note: Revealed type is "builtins.set[builtins.int]"\nSuccess: no issues found in 1 source file\n'  # noqa: E501
            ),
        ),
        (
            fixture_dir("to_dict_decorator.py"),
            fixture_dir(
                'to_dict_decorator.py:17: note: Revealed type is "builtins.dict[builtins.int, builtins.int]"\nSuccess: no issues found in 1 source file\n'  # noqa: E501
            ),
        ),
        (
            fixture_dir("to_ordered_dict_decorator.py"),
            fixture_dir(
                'to_ordered_dict_decorator.py:17: note: Revealed type is "collections.OrderedDict[builtins.int, builtins.int]"\nSuccess: no issues found in 1 source file\n'  # noqa: E501
            ),
        ),
        (
            fixture_dir("apply_to_return_value_decorator.py"),
            fixture_dir(
                'apply_to_return_value_decorator.py:20: note: Revealed type is "builtins.list[builtins.int]"\nSuccess: no issues found in 1 source file\n'  # noqa: E501
            ),
        ),
    ),
)
def test_type_inference(fixture, message):
    check_mypy_run(MYPY_ARGS + [fixture], message)



import pytest

from eth_utils import (
    combomethod,
)


@pytest.fixture()
def Getter():
    class _Getter:
        val = 1

        @combomethod
        def get(combo):
            if isinstance(combo, type):
                return f"{combo.val} by class"
            elif isinstance(combo, _Getter):
                return f"{combo.val} by instance"
            else:
                raise TypeError("Unreachable, unless you really monkey around")

    return _Getter


def test_class_access(Getter):
    assert Getter.get() == "1 by class"


def test_instance_access(Getter):
    getter = Getter()
    getter.val = 2
    assert getter.get() == "2 by instance"



import pytest
from abc import (
    ABCMeta,
    abstractmethod,
)
import logging

from eth_utils import (
    HasLoggerMeta,
)


class HasLoggerCompatWithABC(metaclass=HasLoggerMeta.meta_compat(ABCMeta)):
    @abstractmethod
    def something(self):
        ...


def test_has_logger_compat_with_abc_ABC():
    assert hasattr(HasLoggerCompatWithABC, "logger")
    assert isinstance(HasLoggerCompatWithABC.logger, logging.Logger)
    assert HasLoggerCompatWithABC.logger.name.endswith("HasLoggerCompatWithABC")


def test_abc_still_enforces_abstract_methods():
    with pytest.raises(TypeError):
        HasLoggerCompatWithABC()



import logging
import pickle
import uuid

from eth_utils.logging import (
    ExtendedDebugLogger,
    get_extended_debug_logger,
    get_logger,
)


class CustomLogger(logging.Logger):
    pass


def test_get_logger_with_no_class():
    path = f"testing.{uuid.uuid4()}"
    logger = get_logger(path)
    assert isinstance(logger, logging.Logger)
    assert logger.name == path


def test_get_logger_with_default_class():
    path = f"testing.{uuid.uuid4()}"
    logger = get_logger(path, logging.Logger)
    assert isinstance(logger, logging.Logger)
    assert logger.name == path


def test_get_logger_with_CustomLogger():
    path = f"testing.{uuid.uuid4()}"
    logger = get_logger(path, CustomLogger)
    assert isinstance(logger, CustomLogger)
    assert logger.name == path


def test_get_extended_debug_logger():
    path = f"testing.{uuid.uuid4()}"
    logger = get_extended_debug_logger(path)
    assert isinstance(logger, ExtendedDebugLogger)
    assert logger.name == path


def test_get_extended_debug_logger_if_other_logger_in_cache():
    path = f"testing.{uuid.uuid4()}"
    normal_logger = get_logger(path)
    assert not isinstance(normal_logger, ExtendedDebugLogger)
    assert normal_logger.name == path

    extended_logger = get_extended_debug_logger(path)
    assert isinstance(extended_logger, ExtendedDebugLogger)
    assert extended_logger.name == path


def test_get_logger_preserves_logging_module_config():
    assert logging.getLoggerClass() is logging.Logger

    path = f"testing.{uuid.uuid4()}"
    logger = get_logger(path, CustomLogger)
    assert isinstance(logger, CustomLogger)

    assert logging.getLoggerClass() is logging.Logger


def test_extended_debug_logger_pickling():
    name = f"testing.{uuid.uuid4()}"
    extended_logger = get_extended_debug_logger(name)
    pickled = pickle.dumps(extended_logger)

    # logging.Logger.__reduce__() returns (getLogger, name), which could give us a
    # regular Logger instance after pickling/unpickling an ExtendedDebugLogger
    # (e.g. when crossing process boundaries since the other process wouldn't have an
    # ExtendedDebugLogger with that name in the logging module's cache). This test
    # ensures an unpickled ExtendedDebugLogger is still an ExtendedDebugLogger even
    # when it's not in the logging's module cache.
    del logging.Logger.manager.loggerDict[name]
    unpickled_logger = pickle.loads(pickled)
    assert isinstance(unpickled_logger, ExtendedDebugLogger)



from eth_utils.decorators import (
    combomethod,
)
from eth_utils.module_loading import (
    import_string,
)


def test_import_string():
    imported_combomethod = import_string("eth_utils.decorators.combomethod")
    assert imported_combomethod is combomethod



import pytest
import re
from typing import (
    Any,
    Dict,
    Literal,
    Mapping,
    NamedTuple,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from eth_typing import (
    ABI,
    ABIComponent,
    ABIComponentIndexed,
    ABIConstructor,
    ABIElement,
    ABIError,
    ABIEvent,
    ABIFallback,
    ABIFunction,
    ABIReceive,
    HexStr,
)

from eth_utils.abi import (
    abi_to_signature,
    collapse_if_tuple,
    event_abi_to_log_topic,
    event_signature_to_log_topic,
    filter_abi_by_name,
    filter_abi_by_type,
    function_abi_to_4byte_selector,
    function_signature_to_4byte_selector,
    get_abi_input_names,
    get_abi_input_types,
    get_abi_output_names,
    get_abi_output_types,
    get_aligned_abi_inputs,
    get_all_event_abis,
    get_all_function_abis,
    get_normalized_abi_inputs,
)
from eth_utils.hexadecimal import (
    encode_hex,
)

ABI_FUNCTION_NO_NAME = ABIFunction({"type": "function"})  # type: ignore

ABI_FUNCTION_NO_INPUTS = ABIFunction({"name": "noInputs", "type": "function"})

ABI_FUNCTION_TOKEN_LAUNCHED = ABIFunction(
    {"name": "tokenLaunched", "type": "function", "inputs": []}
)
ABI_FUNCTION_TOKEN_LAUNCHED_WITH_INPUT = ABIFunction(
    {
        "name": "tokenLaunched",
        "type": "function",
        "inputs": [{"type": "int256", "name": "value"}],
    }
)

ABI_FUNCTION_CEILING = ABIFunction(
    {"name": "CEILING", "type": "function", "inputs": []}
)

ABI_FUNCTION_REGISTRAR = ABIFunction(
    {
        "name": "Registrar",
        "type": "function",
        "inputs": [
            {"type": "address", "name": "a"},
            {"type": "bytes32", "name": "b"},
            {"type": "address", "name": "c"},
        ],
    }
)

ABI_FUNCTION_FOUR_NAMED_ARGS = ABIFunction(
    {
        "constant": False,
        "inputs": [
            {"name": "a", "type": "int256"},
            {"name": "b", "type": "int256"},
            {"name": "c", "type": "int256"},
            {"name": "d", "type": "int256"},
        ],
        "name": "fourNamedArgs",
        "outputs": [],
        "type": "function",
    }
)

ABI_FUNCTION_THREE_NAMED_ARGS_DUPLICATE = ABIFunction(
    {
        "constant": False,
        "inputs": [
            {"name": "b", "type": "int256"},
            {"name": "b", "type": "int256"},
            {"name": "a", "type": "int256"},
        ],
        "name": "threeNamedArgs",
        "outputs": [],
        "type": "function",
    }
)

ABI_FUNCTION_TWO_NAMED_ARGS = ABIFunction(
    {
        "type": "function",
        "name": "FnMultiArg",
        "inputs": [
            {"name": "arg0", "type": "uint256"},
            {"name": "arg1", "type": "uint256"},
        ],
        "outputs": [
            {"name": "arg0", "type": "uint256"},
            {"name": "arg1", "type": "uint256"},
        ],
    }
)

ABI_FUNCTION_MULTI_TYPED_ARGS = ABIFunction(
    {
        "type": "function",
        "name": "FnMultiTypedArg",
        "inputs": [
            {"name": "arg0", "type": "uint256"},
            {"name": "arg1", "type": "bytes32"},
        ],
        "outputs": [
            {"name": "arg0", "type": "uint256"},
            {"name": "arg1", "type": "bytes32"},
        ],
    }
)

ABI_FUNCTION_TWO_UNNAMED_ARGS = ABIFunction(
    {
        "constant": False,
        "inputs": [
            {"name": "", "type": "int256"},
            {"name": "", "type": "int256"},
        ],
        "name": "twoUnnamedArgs",
        "outputs": [],
        "type": "function",
    }
)

ABI_FUNCTION_SINGLE_ARG = ABIFunction(
    {
        "type": "function",
        "name": "FnSingleArg",
        "inputs": [{"name": "arg0", "type": "uint256"}],
        "outputs": [{"name": "arg0", "type": "uint256"}],
    }
)

ABI_FUNCTION_ADD = ABIFunction(
    {
        "constant": False,
        "inputs": [
            {"name": "x", "type": "uint256"},
            {"name": "y", "type": "uint256"},
        ],
        "name": "add",
        "outputs": [{"name": "sum", "type": "uint256"}],
        "type": "function",
    }
)

ABI_FUNCTION_TUPLE_INPUT_WITH_OUTPUTS = ABIFunction(
    {
        "inputs": [
            {
                "components": [{"name": "x", "type": "uint256"}],
                "type": "tuple",
                "name": "aTuple",
            }
        ],
        "name": "singletonTupleInput",
        "outputs": [{"name": "y", "type": "uint256"}],
        "type": "function",
    }
)

ABI_FUNCTION_TUPLE_INPUT_WITH_MULTI_TYPE_OUTPUTS = ABIFunction(
    {
        "inputs": [
            {
                "components": [{"name": "x", "type": "uint256"}],
                "type": "tuple",
                "name": "aTuple",
            }
        ],
        "name": "singletonTupleInput",
        "outputs": [
            {"name": "anAddress", "type": "address"},
            {"name": "anInt", "type": "uint256"},
            {"name": "someBytes", "type": "bytes"},
        ],
        "type": "function",
    }
)

ABI_FUNCTION_NESTED_TUPLE_INPUTS = ABIFunction(
    {
        "inputs": [
            {
                "components": [
                    {"name": "anAddress", "type": "address"},
                    {"name": "anInt", "type": "uint256"},
                    {"name": "someBytes", "type": "bytes"},
                    {
                        "name": "aTuple",
                        "type": "tuple",
                        "components": [
                            {"name": "anAddress", "type": "address"},
                            {"name": "anInt", "type": "uint256"},
                            {"name": "someBytes", "type": "bytes"},
                        ],
                    },
                ],
                "type": "tuple",
                "name": "aTuple",
            }
        ],
        "name": "nestedTupleInputs",
        "type": "function",
    }
)

ABI_FUNCTION_ZERO_TUPLE_INPUT = ABIFunction(
    {
        "inputs": [{"components": [], "type": "tuple", "name": "zroTuple"}],
        "name": "zeroTupleInput",
        "type": "function",
    }
)

ABI_FUNCTION_SINGLETON_TUPLE_INPUT = ABIFunction(
    {
        "inputs": [
            {
                "components": [{"name": "anAddress", "type": "address"}],
                "type": "tuple",
                "name": "aTuple",
            }
        ],
        "name": "singletonTupleInput",
        "type": "function",
    }
)

ABI_FUNCTION_ARRAY_OF_TUPLES = ABIFunction(
    {
        "name": "tupleArrayInput",
        "type": "function",
        "inputs": [
            {
                "name": "tupleArray",
                "type": "tuple[]",
                "components": [
                    {"name": "anAddress", "type": "address"},
                    {"name": "anInt", "type": "uint256"},
                    {"name": "someBytes", "type": "bytes"},
                ],
            }
        ],
    }
)

ABI_FUNCTION_FIXED_ARRAY_OF_TUPLES = ABIFunction(
    {
        "name": "tupleFixedArrayInput",
        "type": "function",
        "inputs": [
            {
                "name": "tupleArrayFixed",
                "type": "tuple[5]",
                "components": [
                    {"name": "anAddress", "type": "address"},
                    {"name": "anInt", "type": "uint256"},
                    {"name": "someBytes", "type": "bytes"},
                ],
            }
        ],
    }
)

ABI_FUNCTION_MULTI_TUPLE_INPUTS = ABIFunction(
    {
        "constant": False,
        "inputs": [
            {
                "components": [
                    {"name": "a", "type": "uint256"},
                    {"name": "b", "type": "uint256[]"},
                    {
                        "components": [
                            {"name": "x", "type": "uint256"},
                            {"name": "y", "type": "uint256"},
                        ],
                        "name": "c",
                        "type": "tuple[]",
                    },
                ],
                "name": "s",
                "type": "tuple",
            },
            {
                "components": [
                    {"name": "x", "type": "uint256"},
                    {"name": "y", "type": "uint256"},
                ],
                "name": "t",
                "type": "tuple",
            },
            {"name": "a", "type": "uint256"},
            {
                "components": [
                    {"name": "x", "type": "uint256"},
                    {"name": "y", "type": "uint256"},
                ],
                "name": "b",
                "type": "tuple[][]",
            },
        ],
        "name": "multiTupleInputs",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    }
)

ABI_CONSTRUCTOR = ABIConstructor({"type": "constructor"})

ABI_CONSTRUCTOR_WITH_INPUT = ABIConstructor(
    {
        "inputs": [{"name": "started", "type": "str"}],
        "type": "constructor",
    }
)

ABI_FALLBACK = ABIFallback({"type": "fallback"})

ABI_RECEIVE = ABIReceive({"type": "receive"})

ABI_ERROR = ABIError({"type": "error", "name": "error"})

ABI_ERROR_INVALID = ABIError(
    {
        "name": "Invalid",
        "type": "error",
        "inputs": [
            {"type": "address", "name": "a"},
            {"type": "bytes32", "name": "b"},
        ],
    }
)

# Intentionally invalid ABIError, input names are missing
ABI_ERROR_FAILED = {
    "type": "error",
    "name": "failed",
    "inputs": [
        {"x": b"1", "type": "bytes32"},
        {"y": "value", "type": "str"},
    ],
}

ABI_ERROR_MY_ERROR = ABIError({"type": "error", "name": "myError"})

ABI_EVENT_FINISHED = ABIEvent({"type": "event", "name": "finished", "inputs": []})

ABI_EVENT_FOO = ABIEvent({"type": "event", "name": "foo"})

# Intentionally invalid ABIEvent, missing event name and input names
ABI_EVENT_NO_NAME = {
    "type": "event",
    "inputs": [
        {"x": 1, "type": "int"},
    ],
}

ABI_EVENT_TRANSFER = ABIEvent(
    {
        "anonymous": False,
        "name": "Transfer",
        "type": "event",
        "inputs": [
            ABIComponentIndexed({"indexed": True, "name": "from", "type": "address"}),
            ABIComponentIndexed({"indexed": True, "name": "to", "type": "address"}),
            ABIComponentIndexed({"indexed": False, "name": "value", "type": "uint256"}),
        ],
    }
)

ABI_EVENT_LOG_NO_ARG = ABIEvent(
    {
        "anonymous": False,
        "name": "LogNoArg",
        "type": "event",
        "inputs": [],
    }
)

ABI_EVENT_LOG_SINGLE_ARG = ABIEvent(
    {
        "anonymous": False,
        "name": "LogSingleArg",
        "type": "event",
        "inputs": [{"name": "arg0", "type": "uint256"}],
    }
)

ABI_EVENT_LOG_SINGLE_WITH_INDEX = ABIEvent(
    {
        "anonymous": False,
        "name": "LogSingleWithIndex",
        "type": "event",
        "inputs": [
            ABIComponentIndexed({"indexed": True, "name": "arg0", "type": "uint256"})
        ],
    }
)

ABI_EVENT_LOG_TWO_EVENTS = ABIEvent(
    {
        "anonymous": False,
        "name": "logTwoEvents",
        "type": "event",
        "inputs": [
            {"name": "_arg0", "type": "uint256"},
        ],
    }
)

ABI_EVENT_LOG_MULTI_ARGS = ABIEvent(
    {
        "type": "event",
        "name": "LogMultiArg",
        "inputs": [
            {"name": "arg0", "type": "uint256"},
            {"name": "arg1", "type": "uint256"},
            {"name": "arg2", "type": "uint256"},
        ],
    }
)

ABI_COMPONENT = ABIComponent({"type": "str", "name": "aComponent"})

ABI_COMPONENT_TUPLE = ABIComponent(
    {
        "components": [
            {"name": "anAddress", "type": "address"},
            {"name": "anInt", "type": "uint256"},
            {"name": "someBytes", "type": "bytes"},
        ],
        "type": "tuple",
        "name": "aTuple",
    }
)

ABI_COMPONENT_TUPLE_ARRAY = ABIComponent(
    {
        "components": [
            {"name": "bytes32", "type": "bytes32[]"},
            {"name": "bytes", "type": "bytes"},
            {"name": "someMoarBytes", "type": "bytes"},
        ],
        "type": "tuple[]",
        "name": "aTupleArray",
    }
)

ABI_COMPONENT_MULTI_DIM_TUPLE = ABIComponent(
    {
        "components": [
            {"name": "anAddress", "type": "address"},
            {"name": "anInt", "type": "uint256"},
            {"name": "someBytes", "type": "bytes"},
        ],
        "type": "tuple[3]",
        "name": "aMultiDimensionalTuple",
    }
)


def build_contract_abi(abi_elements: Sequence[ABIElement]) -> ABI:
    return cast(ABI, abi_elements)


@pytest.mark.parametrize(
    "abi_element,expected",
    (
        (ABI_FUNCTION_TOKEN_LAUNCHED, "0xde78e78a"),
        (ABI_FUNCTION_CEILING, "0xc51bf934"),
        (ABI_FUNCTION_REGISTRAR, "0xa31d5580"),
        (ABI_FUNCTION_NO_INPUTS, "0xc1772ae8"),
        (ABI_FUNCTION_NESTED_TUPLE_INPUTS, "0xfa89ea29"),
        (ABI_EVENT_FINISHED, "0xbef4876b"),
        (ABI_EVENT_FOO, "0xc2985578"),
        (ABI_EVENT_NO_NAME, "0x77c874c7"),
        (ABI_FALLBACK, "0x552079dc"),
        (ABI_RECEIVE, "0xa3e76c0f"),
        (ABI_CONSTRUCTOR, "0x90fa17bb"),
        (ABI_CONSTRUCTOR_WITH_INPUT, "0x2757c293"),
        (ABI_ERROR_FAILED, "0x236ac042"),
    ),
)
def test_fn_abi_to_4byte_selector(abi_element: ABIElement, expected: HexStr) -> None:
    bytes_selector = function_abi_to_4byte_selector(abi_element)
    hex_selector = encode_hex(bytes_selector)
    assert hex_selector == expected


@pytest.mark.parametrize(
    "signature,expected",
    (
        ("tokenLaunched()", "0xde78e78a"),
        ("CEILING()", "0xc51bf934"),
        ("Registrar(address,bytes32,address)", "0xa31d5580"),
        ("failed(bytes32,str)", "0x236ac042"),
        ("receive()", "0xa3e76c0f"),
        ("fallback()", "0x552079dc"),
        ("constructor()", "0x90fa17bb"),
    ),
)
def test_fn_signature_to_4byte_selector(signature: str, expected: HexStr) -> None:
    bytes_selector = function_signature_to_4byte_selector(signature)
    hex_selector = encode_hex(bytes_selector)
    assert hex_selector == expected


@pytest.mark.parametrize(
    "abi_element,expected",
    (
        (ABI_FUNCTION_NO_NAME, "function()"),
        (ABI_FUNCTION_NO_INPUTS, "noInputs()"),
        (ABI_FUNCTION_ARRAY_OF_TUPLES, "tupleArrayInput((address,uint256,bytes)[])"),
        (
            ABI_FUNCTION_FIXED_ARRAY_OF_TUPLES,
            "tupleFixedArrayInput((address,uint256,bytes)[5])",
        ),
        (
            ABI_FUNCTION_NESTED_TUPLE_INPUTS,
            "nestedTupleInputs((address,uint256,bytes,(address,uint256,bytes)))",
        ),
        (ABI_FUNCTION_ZERO_TUPLE_INPUT, "zeroTupleInput(())"),
        (ABI_FUNCTION_SINGLETON_TUPLE_INPUT, "singletonTupleInput((address))"),
        (
            ABI_ERROR_FAILED,
            "failed(bytes32,str)",
        ),
        (ABI_ERROR_MY_ERROR, "myError()"),
        (ABI_EVENT_FOO, "foo()"),
        (
            ABI_EVENT_NO_NAME,
            "event(int)",
        ),
        (ABI_EVENT_LOG_MULTI_ARGS, "LogMultiArg(uint256,uint256,uint256)"),
        (ABI_CONSTRUCTOR, "constructor()"),
        (
            ABI_CONSTRUCTOR_WITH_INPUT,
            "constructor(str)",
        ),
        (ABI_FALLBACK, "fallback()"),
        (ABI_RECEIVE, "receive()"),
        (ABI_ERROR, "error()"),
    ),
)
def test_abi_to_signature(abi_element: ABIElement, expected: str) -> None:
    assert abi_to_signature(abi_element) == expected


@pytest.mark.parametrize(
    "abi_elements,expected_event_abis",
    (
        (
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
            [],
        ),
        (
            [
                ABI_EVENT_LOG_NO_ARG,
                ABI_EVENT_LOG_SINGLE_ARG,
                ABI_EVENT_LOG_SINGLE_WITH_INDEX,
            ],
            [
                ABI_EVENT_LOG_NO_ARG,
                ABI_EVENT_LOG_SINGLE_ARG,
                ABI_EVENT_LOG_SINGLE_WITH_INDEX,
            ],
        ),
        (
            [
                ABI_FUNCTION_TOKEN_LAUNCHED,
                ABI_EVENT_LOG_NO_ARG,
                ABI_ERROR,
            ],
            [
                ABI_EVENT_LOG_NO_ARG,
            ],
        ),
        (
            [ABI_FUNCTION_ADD],
            [],
        ),
        (
            [
                ABI_FUNCTION_ADD,
                ABI_FALLBACK,
                ABI_RECEIVE,
                ABI_CONSTRUCTOR_WITH_INPUT,
                ABI_ERROR_FAILED,
            ],
            [],
        ),
        (
            [
                ABI_EVENT_LOG_NO_ARG,
                ABI_FUNCTION_ADD,
                ABI_FALLBACK,
                ABI_RECEIVE,
                ABI_CONSTRUCTOR_WITH_INPUT,
                ABI_EVENT_LOG_SINGLE_WITH_INDEX,
            ],
            [
                ABI_EVENT_LOG_NO_ARG,
                ABI_EVENT_LOG_SINGLE_WITH_INDEX,
            ],
        ),
    ),
)
def test_get_all_event_abis(
    abi_elements: Sequence[ABIElement], expected_event_abis: Sequence[ABIEvent]
) -> None:
    contract_abi = build_contract_abi(abi_elements)
    assert get_all_event_abis(contract_abi) == expected_event_abis


@pytest.mark.parametrize(
    "abi_elements,abi_name,expected_element_abis",
    (
        (
            [ABI_EVENT_LOG_NO_ARG],
            "LogNoArg",
            [ABI_EVENT_LOG_NO_ARG],
        ),
        (
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
            "tokenLaunched",
            [ABI_FUNCTION_TOKEN_LAUNCHED],
        ),
        (
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
            "notafunction",
            [],
        ),
        (
            [ABI_ERROR_INVALID],
            "Invalid",
            [ABI_ERROR_INVALID],
        ),
        (
            [ABI_RECEIVE],  # receive abi has no name
            "receive",
            [],
        ),
        (
            [ABI_FALLBACK],  # fallback abi has no name
            "fallback",
            [],
        ),
        (
            [ABI_CONSTRUCTOR_WITH_INPUT],  # constructor abi has no name
            "constructor",
            [],
        ),
        (
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_TOKEN_LAUNCHED],
            "tokenLaunched",
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_TOKEN_LAUNCHED],
        ),
        (
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_TOKEN_LAUNCHED_WITH_INPUT],
            "tokenLaunched",
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_TOKEN_LAUNCHED_WITH_INPUT],
        ),
    ),
)
def test_filter_abi_by_name(
    abi_elements: Sequence[ABIElement],
    abi_name: str,
    expected_element_abis: Sequence[ABIElement],
) -> None:
    contract_abi = build_contract_abi(abi_elements)
    assert filter_abi_by_name(abi_name, contract_abi) == expected_element_abis


@pytest.mark.parametrize(
    "abi_elements,abi_type,expected_element_abis",
    (
        (
            [
                ABI_EVENT_LOG_TWO_EVENTS,
                ABI_FUNCTION_TOKEN_LAUNCHED,
                ABI_FUNCTION_CEILING,
                ABI_FUNCTION_REGISTRAR,
            ],
            "function",
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
        ),
        (
            [],
            "function",
            [],
        ),
        (
            [
                ABI_EVENT_LOG_TWO_EVENTS,
                ABI_FUNCTION_TOKEN_LAUNCHED,
                ABI_FUNCTION_CEILING,
                ABI_FUNCTION_REGISTRAR,
            ],
            "event",
            [ABI_EVENT_LOG_TWO_EVENTS],
        ),
        (
            [ABI_FALLBACK],
            "fallback",
            [ABI_FALLBACK],
        ),
        (
            [ABI_FUNCTION_TOKEN_LAUNCHED],
            "fallback",
            [],
        ),
        (
            [ABI_RECEIVE, ABI_FALLBACK],
            "receive",
            [ABI_RECEIVE],
        ),
        (
            [ABI_CONSTRUCTOR_WITH_INPUT],
            "constructor",
            [ABI_CONSTRUCTOR_WITH_INPUT],
        ),
        (
            [ABI_ERROR_INVALID],
            "error",
            [ABI_ERROR_INVALID],
        ),
    ),
)
def test_filter_abi_by_type(
    abi_elements: Sequence[ABIElement],
    abi_type: Literal[
        "function", "event", "fallback", "receive", "constructor", "error"
    ],
    expected_element_abis: Sequence[
        Union[ABIFunction, ABIConstructor, ABIFallback, ABIReceive, ABIEvent, ABIError]
    ],
) -> None:
    contract_abi = build_contract_abi(abi_elements)
    actual_element_abis = filter_abi_by_type(abi_type, contract_abi)
    assert actual_element_abis == expected_element_abis
    assert all(abi["type"] == abi_type for abi in actual_element_abis)


def test_filter_abi_by_type_raises_for_invalid_abi_type() -> None:
    contract_abi = build_contract_abi([ABI_FUNCTION_TOKEN_LAUNCHED])
    with pytest.raises(
        ValueError,
        match=re.escape("Unsupported ABI type: typing.Literal['notanabitype']"),
    ):
        filter_abi_by_type(Literal["notanabitype"], contract_abi)  # type: ignore


@pytest.mark.parametrize(
    "abi_element,expected_names,expected_types",
    (
        (
            ABI_EVENT_LOG_SINGLE_ARG,
            ["arg0"],
            ["uint256"],
        ),
        (
            ABI_EVENT_LOG_SINGLE_WITH_INDEX,
            ["arg0"],
            ["uint256"],
        ),
        (
            ABI_EVENT_LOG_MULTI_ARGS,
            ["arg0", "arg1", "arg2"],
            ["uint256", "uint256", "uint256"],
        ),
        (
            ABI_EVENT_LOG_NO_ARG,
            [],
            [],
        ),
        (
            ABI_FUNCTION_SINGLE_ARG,
            ["arg0"],
            ["uint256"],
        ),
        (
            ABI_FUNCTION_TWO_NAMED_ARGS,
            ["arg0", "arg1"],
            ["uint256", "uint256"],
        ),
        (
            ABI_FUNCTION_MULTI_TYPED_ARGS,
            ["arg0", "arg1"],
            ["uint256", "bytes32"],
        ),
        (
            ABI_FUNCTION_TWO_UNNAMED_ARGS,
            ["", ""],
            ["int256", "int256"],
        ),
        (
            ABI_FUNCTION_TUPLE_INPUT_WITH_OUTPUTS,
            ["aTuple"],
            ["(uint256)"],
        ),
        (
            ABI_FUNCTION_TUPLE_INPUT_WITH_MULTI_TYPE_OUTPUTS,
            ["aTuple"],
            ["(uint256)"],
        ),
        (
            ABI_CONSTRUCTOR_WITH_INPUT,
            ["started"],
            ["str"],
        ),
        (
            ABI_FUNCTION_NO_INPUTS,
            [],
            [],
        ),
        (
            ABI_ERROR_INVALID,
            ["a", "b"],
            ["address", "bytes32"],
        ),
    ),
)
def test_get_input_names_and_types_from_abi_element(
    abi_element: ABIElement,
    expected_names: Sequence[str],
    expected_types: Sequence[str],
) -> None:
    assert get_abi_input_names(abi_element) == expected_names
    assert get_abi_input_types(abi_element) == expected_types


@pytest.mark.parametrize(
    "abi_element",
    (
        ABI_FALLBACK,
        ABI_RECEIVE,
    ),
)
def test_get_input_names_and_types_raises_for_fallback_or_receive_abi(
    abi_element: ABIElement,
) -> None:
    with pytest.raises(
        ValueError,
        match=f"Inputs not supported for function types `fallback` or `receive`."
        f" Provided ABI type was `{abi_element['type']}` with inputs"
        f" `None`.",
    ):
        get_abi_input_names(abi_element)

    with pytest.raises(
        ValueError,
        match=f"Inputs not supported for function types `fallback` or `receive`."
        f" Provided ABI type was `{abi_element['type']}` with inputs"
        f" `None`.",
    ):
        get_abi_input_types(abi_element)


@pytest.mark.parametrize(
    "abi_element,expected_names,expected_types",
    [
        (
            ABI_FUNCTION_SINGLE_ARG,
            ["arg0"],
            ["uint256"],
        ),
        (
            ABI_FUNCTION_TWO_NAMED_ARGS,
            ["arg0", "arg1"],
            ["uint256", "uint256"],
        ),
        (
            ABI_FUNCTION_MULTI_TYPED_ARGS,
            ["arg0", "arg1"],
            ["uint256", "bytes32"],
        ),
        (
            ABI_FUNCTION_TWO_UNNAMED_ARGS,
            [],
            [],
        ),
        (
            ABI_FUNCTION_NO_INPUTS,
            [],
            [],
        ),
        (
            ABI_FUNCTION_TUPLE_INPUT_WITH_OUTPUTS,
            ["y"],
            ["uint256"],
        ),
        (
            ABI_FUNCTION_TUPLE_INPUT_WITH_MULTI_TYPE_OUTPUTS,
            ["anAddress", "anInt", "someBytes"],
            ["address", "uint256", "bytes"],
        ),
    ],
)
def test_get_abi_output_names(
    abi_element: ABIElement,
    expected_names: Sequence[str],
    expected_types: Sequence[str],
) -> None:
    assert get_abi_output_names(abi_element) == expected_names
    assert get_abi_output_types(abi_element) == expected_types


@pytest.mark.parametrize(
    "abi_element",
    (
        ABI_CONSTRUCTOR,
        ABI_CONSTRUCTOR_WITH_INPUT,
        ABI_EVENT_FINISHED,
        ABI_EVENT_LOG_NO_ARG,
        ABI_ERROR,
        ABI_ERROR_INVALID,
        ABI_FALLBACK,
        ABI_RECEIVE,
    ),
)
def test_get_abi_output_names_raises_for_non_function_types(
    abi_element: ABIElement,
) -> None:
    with pytest.raises(
        ValueError,
        match=f"Outputs only supported for ABI type `function`."
        f" Provided ABI type was `{abi_element['type']}` and outputs were `None`.",
    ):
        get_abi_output_names(abi_element)

    with pytest.raises(
        ValueError,
        match=f"Outputs only supported for ABI type `function`."
        f" Provided ABI type was `{abi_element['type']}` and outputs were `None`.",
    ):
        get_abi_output_types(abi_element)


@pytest.mark.parametrize(
    "abi_elements,expected_function_abis",
    (
        (
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
        ),
        (
            [
                ABI_FUNCTION_TOKEN_LAUNCHED,
                ABI_FUNCTION_CEILING,
                ABI_FUNCTION_REGISTRAR,
                ABI_EVENT_LOG_NO_ARG,
            ],
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
        ),
        (
            [
                ABI_FUNCTION_TOKEN_LAUNCHED,
                ABI_FUNCTION_CEILING,
                ABI_FUNCTION_REGISTRAR,
                ABI_ERROR_INVALID,
            ],
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
        ),
        (
            [
                ABI_FUNCTION_TOKEN_LAUNCHED,
                ABI_FUNCTION_CEILING,
                ABI_FUNCTION_REGISTRAR,
                ABI_FALLBACK,
            ],
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
        ),
        (
            [
                ABI_FUNCTION_TOKEN_LAUNCHED,
                ABI_FUNCTION_CEILING,
                ABI_FUNCTION_REGISTRAR,
                ABI_RECEIVE,
                ABI_ERROR,
            ],
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
        ),
        (
            [
                ABI_FUNCTION_TOKEN_LAUNCHED,
                ABI_FUNCTION_CEILING,
                ABI_FUNCTION_REGISTRAR,
                ABI_CONSTRUCTOR_WITH_INPUT,
            ],
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
        ),
        (
            [
                ABI_FUNCTION_TOKEN_LAUNCHED,
                ABI_FUNCTION_CEILING,
                ABI_FUNCTION_REGISTRAR,
                ABI_FALLBACK,
                ABI_RECEIVE,
                ABI_CONSTRUCTOR_WITH_INPUT,
            ],
            [ABI_FUNCTION_TOKEN_LAUNCHED, ABI_FUNCTION_CEILING, ABI_FUNCTION_REGISTRAR],
        ),
        (
            [ABI_FALLBACK, ABI_RECEIVE, ABI_CONSTRUCTOR_WITH_INPUT, ABI_ERROR_INVALID],
            [],
        ),
    ),
)
def test_get_all_function_abis(
    abi_elements: Sequence[ABIElement], expected_function_abis: Sequence[ABIFunction]
) -> None:
    contract_abi = build_contract_abi(abi_elements)
    function_abis = get_all_function_abis(contract_abi)
    assert function_abis == expected_function_abis


@pytest.mark.parametrize(
    "abi_event,expected",
    (
        (
            ABI_EVENT_TRANSFER,
            "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef",
        ),
        (
            ABI_EVENT_FINISHED,
            "0xbef4876bcdde2983a1ba9fe22b5983b09081cf602d266f72a613d0d9c2f78e28",
        ),
        (
            ABI_EVENT_LOG_SINGLE_ARG,
            "0x56d2ef3c5228bf5d88573621e325a4672ab50e033749a601e4f4a5e1dce905d4",
        ),
        (
            ABI_EVENT_NO_NAME,
            "0x77c874c798a153cb8bb91a65910a8f49033a603329228caa4b3b1bb4bb5b9e18",
        ),
    ),
)
def test_event_abi_to_log_topic(abi_event: ABIEvent, expected: HexStr) -> None:
    bytes_topic = event_abi_to_log_topic(abi_event)
    hex_topic = encode_hex(bytes_topic)
    assert hex_topic == expected


@pytest.mark.parametrize(
    "event_signature,expected",
    (
        (
            "Transfer(address,address,uint256)",
            "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef",
        ),
        (
            "Event_2(bytes32,uint256)",
            "0x115494e95409cd09eaa4590f66d8938bb342c170bc2edcca5fd14ad46fb7dcb8",
        ),
        (
            "event()",
            "0x4980500b48d9072058dce7e5222cd23e11759678f8aa74b3a6f0e891bb888979",
        ),
    ),
)
def test_event_signature_to_log_topic(event_signature: str, expected: HexStr) -> None:
    bytes_topic = event_signature_to_log_topic(event_signature)
    hex_topic = encode_hex(bytes_topic)
    assert hex_topic == expected


@pytest.mark.parametrize(
    "abi_component,expected_type_signature",
    (
        (
            ABI_COMPONENT_TUPLE,
            "(address,uint256,bytes)",
        ),
        (
            ABI_COMPONENT_TUPLE_ARRAY,
            "(bytes32[],bytes,bytes)[]",
        ),
        (
            ABI_COMPONENT_MULTI_DIM_TUPLE,
            "(address,uint256,bytes)[3]",
        ),
        (
            ABI_COMPONENT,
            "str",
        ),
        (
            (
                {
                    "type": "uint256",
                }
            ),
            "uint256",
        ),
        (
            "uint256",
            "uint256",
        ),
    ),
)
def test_collapse_if_tuple(
    abi_component: Union[ABIComponent, Dict[str, Any], str],
    expected_type_signature: str,
) -> None:
    assert collapse_if_tuple(abi_component) == expected_type_signature


@pytest.mark.parametrize(
    "abi_component,message",
    (
        (
            {
                "type": False,
            },
            "The 'type' must be a string, but got False of type <class 'bool'>",
        ),
        (
            {
                "type": "tuple",
                "components": [{"type": 1}],
            },
            "The 'type' must be a string, but got 1 of type <class 'int'>",
        ),
        (
            {"name": "notype"},
            "The 'type' must be a string, but got None of type <class 'NoneType'>",
        ),
    ),
)
def test_collapse_if_tuple_raises_for_invalid_component(
    abi_component: Union[ABIComponent, Dict[str, Any], str],
    message: str,
) -> None:
    with pytest.raises(TypeError, match=re.escape(message)):
        collapse_if_tuple(abi_component)


@pytest.mark.parametrize(
    "abi_element,args,kwargs,expected",
    (
        (
            ABI_FUNCTION_REGISTRAR,
            (
                "0x1234567890123456789012345678901234567890",
                b"bar",
                "0x1234567890123456789012345678901234567890",
            ),
            {},
            (
                "0x1234567890123456789012345678901234567890",
                b"bar",
                "0x1234567890123456789012345678901234567890",
            ),
        ),
        (
            ABI_FUNCTION_REGISTRAR,
            [],
            {
                "a": "0x1234567890123456789012345678901234567890",
                "b": b"bar",
                "c": "0x1234567890123456789012345678901234567890",
            },
            (
                "0x1234567890123456789012345678901234567890",
                b"bar",
                "0x1234567890123456789012345678901234567890",
            ),
        ),
        (
            ABI_FUNCTION_REGISTRAR,
            [1, "0x1234567890123456789012345678901234567890", b"bar"],
            {},
            (
                1,
                "0x1234567890123456789012345678901234567890",
                b"bar",
            ),
        ),
        (
            ABI_FUNCTION_TWO_UNNAMED_ARGS,
            (1, 2),
            {},
            (1, 2),
        ),
        (
            ABI_FUNCTION_THREE_NAMED_ARGS_DUPLICATE,
            (1, 2, 3),
            {},
            (1, 2, 3),
        ),
        (
            ABI_ERROR_INVALID,
            (
                "0x1234567890123456789012345678901234567890",
                b"bar",
            ),
            {},
            (
                "0x1234567890123456789012345678901234567890",
                b"bar",
            ),
        ),
        (
            ABI_EVENT_LOG_TWO_EVENTS,
            (1,),
            {},
            (1,),
        ),
        (
            ABI_FUNCTION_TOKEN_LAUNCHED,
            (),
            {},
            (),
        ),
        (
            ABI_FUNCTION_NO_INPUTS,
            (),
            {},
            (),
        ),
    ),
)
def test_get_normalized_abi_inputs(
    abi_element: ABIElement,
    args: Sequence[Any],
    kwargs: Dict[str, Any],
    expected: Tuple[Any, ...],
) -> None:
    assert get_normalized_abi_inputs(abi_element, *args, **kwargs) == expected


@pytest.mark.parametrize(
    "abi_element,args,kwargs,error_type,message",
    (
        (
            ABI_FALLBACK,
            (),
            {},
            ValueError,
            "Inputs not supported for function types `fallback` or `receive`. "
            "Provided ABI type was `fallback` with inputs `None`.",
        ),
        (
            ABI_RECEIVE,
            (),
            {},
            ValueError,
            "Inputs not supported for function types `fallback` or `receive`. "
            "Provided ABI type was `receive` with inputs `None`.",
        ),
        (
            ABI_FUNCTION_THREE_NAMED_ARGS_DUPLICATE,
            (1,),
            {"a": 2, "b": 3},
            TypeError,
            "threeNamedArgs() got multiple values for argument(s) 'b'",
        ),
        (
            ABI_FUNCTION_THREE_NAMED_ARGS_DUPLICATE,
            (1,),
            {"d": 2, "e": 3},
            TypeError,
            "threeNamedArgs() got unexpected keyword argument(s) 'd, e",
        ),
        (
            ABI_FUNCTION_TWO_UNNAMED_ARGS,
            tuple(),
            {"x": 1, "y": 2},
            TypeError,
            "twoUnnamedArgs() got unexpected keyword argument(s) 'x, y'",
        ),
        (
            ABI_FUNCTION_TWO_UNNAMED_ARGS,
            (
                1,
                2,
            ),
            {"a": 1, "b": 2},
            TypeError,
            "Incorrect argument count. Expected '2', got '4'.",
        ),
        (
            ABI_ERROR_INVALID,
            (),
            {},
            TypeError,
            "Incorrect argument count. Expected '2', got '0'.",
        ),
    ),
)
def test_get_normalized_abi_inputs_raises_for_invalid_arguments(
    abi_element: ABIElement,
    args: Sequence[Any],
    kwargs: Dict[str, Any],
    error_type: Type[Exception],
    message: str,
) -> None:
    with pytest.raises(
        error_type,
        match=re.escape(message),
    ):
        get_normalized_abi_inputs(abi_element, *args, **kwargs)


class MyXYTuple(NamedTuple):
    x: int
    y: int


GET_ABI_INPUTS_OUTPUT = (
    (
        "(uint256,uint256[],(uint256,uint256)[])",  # Type of s
        "(uint256,uint256)",  # Type of t
        "uint256",  # Type of a
        "(uint256,uint256)[][]",  # Type of b
    ),
    (
        (1, [2, 3, 4], [(5, 6), (7, 8), (9, 10)]),  # Value for s
        (11, 12),  # Value for t
        13,  # Value for a
        [[(14, 15), (16, 17)], [(18, 19)]],  # Value for b
    ),
)


@pytest.mark.parametrize(
    "abi_element,args,expected",
    (
        (
            ABI_FUNCTION_REGISTRAR,
            {
                "a": 1,
                "b": 2,
                "c": 3,
            },  # mapping of arguments
            (
                ("address", "bytes32", "address"),
                (1, 2, 3),
            ),
        ),
        (
            ABI_FUNCTION_REGISTRAR,
            (1, 2, 3),  # tuple of arguments
            (
                ("address", "bytes32", "address"),
                (1, 2, 3),
            ),
        ),
        (
            ABI_FUNCTION_MULTI_TUPLE_INPUTS,
            {
                "s": {
                    "a": 1,
                    "b": [2, 3, 4],
                    "c": [{"x": 5, "y": 6}, {"x": 7, "y": 8}, {"x": 9, "y": 10}],
                },
                "t": {"x": 11, "y": 12},
                "a": 13,
                "b": [[{"x": 14, "y": 15}, {"x": 16, "y": 17}], [{"x": 18, "y": 19}]],
            },
            GET_ABI_INPUTS_OUTPUT,
        ),
        (
            ABI_FUNCTION_MULTI_TUPLE_INPUTS,
            {
                "s": {"a": 1, "b": [2, 3, 4], "c": [(5, 6), (7, 8), {"x": 9, "y": 10}]},
                "t": {"x": 11, "y": 12},
                "a": 13,
                "b": [[(14, 15), (16, 17)], [{"x": 18, "y": 19}]],
            },
            GET_ABI_INPUTS_OUTPUT,
        ),
        (
            ABI_FUNCTION_MULTI_TUPLE_INPUTS,
            {
                "s": {"a": 1, "b": [2, 3, 4], "c": [(5, 6), (7, 8), (9, 10)]},
                "t": (11, 12),
                "a": 13,
                "b": [[(14, 15), (16, 17)], [(18, 19)]],
            },
            GET_ABI_INPUTS_OUTPUT,
        ),
        (
            ABI_FUNCTION_MULTI_TUPLE_INPUTS,
            {
                "s": (1, [2, 3, 4], [(5, 6), (7, 8), (9, 10)]),
                "t": (11, 12),
                "a": 13,
                "b": [[(14, 15), (16, 17)], [(18, 19)]],
            },
            GET_ABI_INPUTS_OUTPUT,
        ),
        (
            ABI_FUNCTION_MULTI_TUPLE_INPUTS,
            (
                (1, [2, 3, 4], [(5, 6), (7, 8), (9, 10)]),
                (11, 12),
                13,
                [[(14, 15), (16, 17)], [(18, 19)]],
            ),
            GET_ABI_INPUTS_OUTPUT,
        ),
        (
            ABI_FUNCTION_MULTI_TUPLE_INPUTS,
            {
                "s": {
                    "a": 1,
                    "b": [2, 3, 4],
                    "c": [(5, 6), (7, 8), MyXYTuple(x=9, y=10)],
                },
                "t": MyXYTuple(x=11, y=12),
                "a": 13,
                "b": [
                    [MyXYTuple(x=14, y=15), MyXYTuple(x=16, y=17)],
                    [MyXYTuple(x=18, y=19)],
                ],
            },
            GET_ABI_INPUTS_OUTPUT,
        ),
        (
            ABI_EVENT_LOG_SINGLE_ARG,
            (1,),
            (("uint256",), (1,)),
        ),
        (
            ABI_ERROR_INVALID,
            ("0x1234567890123456789012345678901234567890", b"bar"),
            (
                ("address", "bytes32"),
                ("0x1234567890123456789012345678901234567890", b"bar"),
            ),
        ),
        (
            ABI_FUNCTION_NO_INPUTS,
            (),
            (
                (),
                (),
            ),
        ),
    ),
)
def test_get_aligned_abi_inputs(
    abi_element: ABIElement,
    args: Union[Tuple[Any, ...], Mapping[Any, Any]],
    expected: Tuple[Tuple[str, ...], Tuple[Any, ...]],
) -> None:
    assert get_aligned_abi_inputs(abi_element, args) == expected


@pytest.mark.parametrize(
    "abi_element,args,error_type,message",
    (
        (
            # raise TypeError, expects list of tuples for input 's.c'
            ABI_FUNCTION_MULTI_TUPLE_INPUTS,
            {
                "s": {"a": 1, "b": [2, 3, 4], "c": ["56", (7, 8), (9, 10)]},
                "t": (11, 12),
                "a": 13,
                "b": [[(14, 15), (16, 17)], [(18, 19)]],
            },
            TypeError,
            'Expected non-string sequence for "tuple" component type: got 56',
        ),
        (
            # raise TypeError, expects list of tuples for input 's.c'
            ABI_FUNCTION_MULTI_TUPLE_INPUTS,
            {
                "s": {"a": 1, "b": [2, 3, 4], "c": {(5, 6), (7, 8), (9, 10)}},
                "t": (11, 12),
                "a": 13,
                "b": [[(14, 15), (16, 17)], [(18, 19)]],
            },
            TypeError,
            'Expected non-string sequence for "tuple[]" component type: got ',
        ),
        (
            ABI_FALLBACK,
            (),
            ValueError,
            "Inputs not supported for function types `fallback` or `receive`. "
            "Provided ABI type was `fallback` with inputs `None`.",
        ),
        (
            ABI_RECEIVE,
            (),
            ValueError,
            "Inputs not supported for function types `fallback` or `receive`. "
            "Provided ABI type was `receive` with inputs `None`.",
        ),
    ),
)
def test_get_aligned_abi_inputs_raises_type_error_for_incorrect_input_types(
    abi_element: ABIElement,
    args: Union[Tuple[Any, ...], Mapping[Any, Any]],
    error_type: Type[Exception],
    message: str,
) -> None:
    with pytest.raises(
        error_type,
        match=re.escape(message),
    ):
        get_aligned_abi_inputs(abi_element, args)



import pytest

from eth_utils import (
    humanize_seconds,
)

SECOND = 1
MINUTE = 60
HOUR = 60 * 60
DAY = 24 * HOUR
YEAR = 365 * DAY
MONTH = YEAR // 12
WEEK = 7 * DAY


@pytest.mark.parametrize(
    "seconds,expected",
    (
        (0, "0s"),
        (1, "1s"),
        (59, "59s"),
        (60, "1m"),
        (61, "1m1s"),
        (119, "1m59s"),
        (59 * MINUTE + 59, "59m59s"),
        (HOUR, "1h"),
        (HOUR + 1, "1h0m1s"),
        (HOUR + MINUTE + 1, "1h1m1s"),
        (HOUR + MINUTE, "1h1m"),
        (23 * HOUR + 59 * MINUTE, "23h59m"),
        (23 * HOUR + 59 * MINUTE + 1, "23h59m1s"),
        (23 * HOUR + 59 * MINUTE + 59, "23h59m59s"),
        (DAY + HOUR, "1d1h"),
        (DAY + HOUR + MINUTE, "1d1h1m"),
        (DAY + MINUTE, "1d0h1m"),
        (DAY + MINUTE + 1, "1d0h1m"),
        (DAY + SECOND, "1d0h0m"),
        (WEEK + DAY + HOUR, "1w1d1h"),
        (WEEK + DAY + HOUR + MINUTE, "1w1d1h"),
        (WEEK + DAY + HOUR + SECOND, "1w1d1h"),
        (WEEK + SECOND, "1w0d0h"),
        (MONTH + WEEK + DAY, "1m1w1d"),
        (MONTH + WEEK + DAY + HOUR, "1m1w1d"),
        (MONTH + SECOND, "1m0w0d"),
        (YEAR + MONTH + WEEK, "1y1m1w"),
        (YEAR + MONTH + WEEK + DAY, "1y1m1w"),
        (YEAR - SECOND, "11m4w2d"),
        (YEAR + SECOND, "1y0m0w"),
    ),
)
def test_humanize_seconds(seconds, expected):
    actual = humanize_seconds(seconds)
    assert actual == expected



from typing import (
    Any,
    Dict,
    Type,
)

from pydantic import (
    BaseModel,
    ConfigDict,
)
from pydantic._internal._core_utils import (
    CoreSchemaField,
)
from pydantic.alias_generators import (
    to_camel,
)
from pydantic.json_schema import (
    DEFAULT_REF_TEMPLATE,
    GenerateJsonSchema,
    JsonSchemaMode,
)


class OmitJsonSchema(GenerateJsonSchema):
    """
    Custom JSON schema generator that omits the schema generation for fields that are
    invalid. Excluded fields (``Field(exclude=True)``) are generally useful as
    properties of the model but are not meant to be serialized to JSON.
    """

    def field_is_present(self, field: CoreSchemaField) -> bool:
        # override ``field_is_present`` and omit excluded fields from the schema
        if field.get("serialization_exclude", False):
            return False
        return super().field_is_present(field)


class CamelModel(BaseModel):
    """
    Camel-case pydantic model. This model is used to ensure serialization in a
    consistent manner, aliasing as camelCase serialization. This is useful for models
    that are used in JSON-RPC requests and responses, marking useful fields for the
    model, but that are not part of the JSON-RPC object, with ``Field(exclude=True)``.
    To serialize a model to the expected JSON-RPC format, or camelCase, use
    ``model_dump(by_alias=True)``.

    .. code-block:: python

        >>> from eth_utils.pydantic import CamelModel
        >>> from pydantic import Field

        >>> class SignedSetCodeAuthorization(CamelModel):
        ...     chain_id: int
        ...     address: bytes
        ...     nonce: int
        ...
        ...     # useful fields for the object but excluded from serialization
        ...     # (not part of the JSON-RPC object)
        ...     authorization_hash: bytes = Field(exclude=True)
        ...     signature: bytes = Field(exclude=True)

        >>> auth = SignedSetCodeAuthorization(
        ...     chain_id=1,
        ...     address=b"0x0000000000000000000000000000000000000000",
        ...     nonce=0,
        ...     authorization_hash=generated_hash,
        ...     signature=generated_signature,
        ... )
        >>> auth.model_dump(by_alias=True)
        {'chainId': 1, 'address': '0x000000000000000000000000000000000000', 'nonce': 0}
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        # populate by snake_case (python) args
        populate_by_name=True,
        # serialize by camelCase (json-rpc) keys
        alias_generator=to_camel,
        # validate default values
        validate_default=True,
    )

    @classmethod
    def model_json_schema(
        cls,
        by_alias: bool = True,
        ref_template: str = DEFAULT_REF_TEMPLATE,
        # default to ``OmitJsonSchema`` to prevent errors from excluded fields
        schema_generator: Type[GenerateJsonSchema] = OmitJsonSchema,
        mode: JsonSchemaMode = "validation",
    ) -> Dict[str, Any]:
        """
        Omits excluded fields from the JSON schema, preventing errors that would
        otherwise be raised by the default schema generator.
        """
        return super().model_json_schema(
            by_alias=by_alias,
            ref_template=ref_template,
            schema_generator=schema_generator,
            mode=mode,
        )
